import sys
import pytest
import numpy as onp
import jax.numpy as jnp
import jax
from numpy.testing import assert_equal, assert_array_equal, assert_

sys.path.insert(0, "../gpax/")

from gpax.models.dkl import DKL
from gpax.utils import get_keys


def get_dummy_data(jax_ndarray=True):
    X = onp.random.randn(10, 36)
    y = onp.random.randn(10,)
    if jax_ndarray:
        return jnp.array(X), jnp.array(y)
    return X, y


@pytest.mark.parametrize("jax_ndarray", [True, False])
def test_fit(jax_ndarray):
    X, y = get_dummy_data(jax_ndarray)
    rng_key = get_keys()[0]
    m = DKL(X.shape[-1])
    m.fit(rng_key, X, y, num_warmup=5, num_samples=5)
    assert m.mcmc is not None


def test_get_mvn_posterior():
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    params = {"w1": jax.random.normal(rng_key, shape=(36, 64)),
              "w2": jax.random.normal(rng_key, shape=(64, 32)),
              "w3": jax.random.normal(rng_key, shape=(32, 2)),
              "b1": jax.random.normal(rng_key, shape=(64,)),
              "b2": jax.random.normal(rng_key, shape=(32,)),
              "b3": jax.random.normal(rng_key, shape=(2,)),
              "k_length": jnp.array([1.0]),
              "k_scale": jnp.array(1.0),
              "noise": jnp.array(0.1)}
    m = DKL(X.shape[-1], kernel='RBF')
    mean, cov = m._get_mvn_posterior(X, y, X_test, params)
    assert isinstance(mean, jnp.ndarray)
    assert isinstance(cov, jnp.ndarray)
    assert_equal(mean.shape, (X_test.shape[0],))
    assert_equal(cov.shape, (X_test.shape[0], X_test.shape[0]))

def test_get_mvn_posterior_noiseless():
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    params = {"w1": jax.random.normal(rng_key, shape=(36, 64)),
              "w2": jax.random.normal(rng_key, shape=(64, 32)),
              "w3": jax.random.normal(rng_key, shape=(32, 2)),
              "b1": jax.random.normal(rng_key, shape=(64,)),
              "b2": jax.random.normal(rng_key, shape=(32,)),
              "b3": jax.random.normal(rng_key, shape=(2,)),
              "k_length": jnp.array([1.0]),
              "k_scale": jnp.array(1.0),
              "noise": jnp.array(0.1)}
    m = DKL(X.shape[-1], kernel='RBF')
    mean1, cov1 = m._get_mvn_posterior(X, y, X_test, params, noiseless=False)
    mean1_, cov1_ = m._get_mvn_posterior(X, y, X_test, params, noiseless=False)
    mean2, cov2 = m._get_mvn_posterior(X, y, X_test, params, noiseless=True)
    assert_array_equal(mean1, mean1_)
    assert_array_equal(cov1, cov1_)
    assert_array_equal(mean1, mean2)
    assert onp.count_nonzero(cov1 - cov2) > 0


def test_jitter_fit():
    rng_key, _ = get_keys()
    X, y = get_dummy_data()
    m = DKL(X.shape[-1], 2, 'RBF')
    m.fit(rng_key, X, y, num_samples=5, num_warmup=5, jitter=1e-6)
    samples1 = m.get_samples()
    m = DKL(X.shape[-1], 2, 'RBF')
    m.fit(rng_key, X, y, num_samples=5, num_warmup=5, jitter=1e-6)
    samples1a = m.get_samples()
    m = DKL(X.shape[-1], 2, 'RBF')
    m.fit(rng_key, X, y, num_samples=5, num_warmup=5, jitter=1e-5)
    samples2 = m.get_samples()
    assert_(onp.count_nonzero(samples1["k_length"] - samples1a["k_length"]) == 0)
    assert_(onp.count_nonzero(samples1["k_length"] - samples2["k_length"]) > 0)


def test_jitter_mvn_posterior():
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    # X = X[None]
    # y = y[None]
    # X_test = X_test[None]
    params = {"w1": jax.random.normal(rng_key, shape=(36, 64)),
              "w2": jax.random.normal(rng_key, shape=(64, 32)),
              "w3": jax.random.normal(rng_key, shape=(32, 2)),
              "b1": jax.random.normal(rng_key, shape=(64,)),
              "b2": jax.random.normal(rng_key, shape=(32,)),
              "b3": jax.random.normal(rng_key, shape=(2,)),
              "k_length": jnp.array([1.0]),
              "k_scale": jnp.array(1.0),
              "noise": jnp.array(0.1)}
    m = DKL(X.shape[-1], 2, 'RBF')
    m.X_train = X
    m.y_train = y
    mean1, cov1 = m._get_mvn_posterior(X, y, X_test, params, jitter=1e-6)
    mean2, cov2 = m._get_mvn_posterior(X, y, X_test, params, jitter=1e-5)
    assert_(onp.count_nonzero(mean1 - mean2) > 0)
    assert_(onp.count_nonzero(cov1 - cov2) > 0)



import sys
import pytest
import numpy as onp
import jax.numpy as jnp
from numpy.testing import assert_, assert_equal

sys.path.insert(0, "../gpax/")

from gpax.models.vi_mtdkl import viMTDKL
from gpax.utils import get_keys


def get_dummy_data():
    X = onp.random.randn(21, 36)
    y = onp.random.randn(21,)
    return jnp.array(X), jnp.array(y)


def attach_indices(X, num_tasks):
    indices = onp.random.randint(0, num_tasks, size=len(X))
    return jnp.column_stack([X, indices])


@pytest.mark.parametrize("num_latents", [1, 2])
@pytest.mark.parametrize("num_tasks", [2, 3])
@pytest.mark.parametrize("data_kernel", ['RBF', 'Matern'])
def test_fit_multitask(data_kernel, num_tasks, num_latents):
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    X = attach_indices(X, num_tasks)
    m = viMTDKL(X.shape[-1] - 1, 2, data_kernel, num_latents=num_latents, shared_input_space=False)
    m.fit(rng_key, X, y, num_steps=10)
    assert_(isinstance(m.kernel_params, dict))
    assert_(isinstance(m.nn_params, dict))


@pytest.mark.parametrize("num_latents", [1, 2])
@pytest.mark.parametrize("num_tasks", [2, 3])
@pytest.mark.parametrize("data_kernel", ['RBF', 'Matern'])
def test_fit_multitask_shared_input(data_kernel, num_tasks, num_latents):
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    y = jnp.repeat(y[:, None], num_tasks, axis=1).reshape(-1)
    m = viMTDKL(X.shape[-1], 2, data_kernel, num_latents=num_latents,
                shared_input_space=True, num_tasks=num_tasks)
    m.fit(rng_key, X, y, num_steps=10)
    assert_(isinstance(m.kernel_params, dict))
    assert_(isinstance(m.nn_params, dict))


@pytest.mark.parametrize("num_latents", [1, 2])
@pytest.mark.parametrize("num_tasks", [2, 3])
@pytest.mark.parametrize("data_kernel", ['RBF', 'Matern'])
def test_fit_predict_multitask(data_kernel, num_tasks, num_latents):
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    X = attach_indices(X, num_tasks)
    m = viMTDKL(X.shape[-1] - 1, 2, data_kernel, num_latents=num_latents, shared_input_space=False)
    m.fit(rng_key, X, y, num_steps=10)
    X_test, _ = get_dummy_data()
    X_test = jnp.column_stack([X_test, jnp.ones(len(X_test))])
    mean, var = m.predict(rng_key, X_test)
    assert_(isinstance(mean, jnp.ndarray))
    assert_(isinstance(var, jnp.ndarray))
    assert_equal(len(mean), len(X_test))
    assert_equal(len(var), len(X_test))



import sys
import pytest
import numpy as onp
import jax.numpy as jnp
import numpyro
from numpy.testing import assert_

sys.path.insert(0, "../gpax/")

from gpax.models.mtgp import MultiTaskGP
from gpax.utils import get_keys


def get_dummy_data():
    X = onp.linspace(1, 2, 20) + 0.1 * onp.random.randn(20,)
    y = (10 * X**2)
    return jnp.array(X), jnp.array(y)


def attach_indices(X, num_tasks):
    indices = onp.random.randint(0, num_tasks, size=len(X))
    return onp.column_stack([X, indices])


def dummy_mean_fn(x, params):
    return params["a"] * x[:, :-1]**params["b"]


def dummy_mean_fn_priors():
    a = numpyro.sample("a", numpyro.distributions.LogNormal(0, 1))
    b = numpyro.sample("b", numpyro.distributions.Normal(3, 1))
    return {"a": a, "b": b}


@pytest.mark.parametrize("num_latents", [1, 2])
@pytest.mark.parametrize("num_tasks", [2, 3])
@pytest.mark.parametrize("data_kernel", ['RBF', 'Matern', 'Periodic'])
def test_fit_multitask(data_kernel, num_tasks, num_latents):
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    X = attach_indices(X, num_tasks)
    m = MultiTaskGP(1, data_kernel, num_latents=num_latents, shared_input_space=False)
    m.fit(rng_key, X, y, num_warmup=50, num_samples=50)
    assert_(isinstance(m.get_samples(), dict))


@pytest.mark.parametrize("num_latents", [1, 2])
@pytest.mark.parametrize("num_tasks", [2, 3])
@pytest.mark.parametrize("data_kernel", ['RBF', 'Matern', 'Periodic'])
def test_fit_multivariate(data_kernel, num_tasks, num_latents):
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    y = jnp.repeat(y[:, None], num_tasks, axis=1).reshape(-1)
    m = MultiTaskGP(
        1, data_kernel, num_latents=num_latents,
        num_tasks=num_tasks, shared_input_space=True)
    m.fit(rng_key, X, y, num_warmup=50, num_samples=50)
    assert_(isinstance(m.get_samples(), dict))


def test_fit_multitask_meanfn():
    rng_key = get_keys()[0]
    X, y = get_dummy_data()
    X = attach_indices(X, 2)
    m = MultiTaskGP(1, 'Matern', num_latents=2, shared_input_space=False,
                    mean_fn=dummy_mean_fn, mean_fn_prior=dummy_mean_fn_priors)
    m.fit(rng_key, X, y, num_warmup=50, num_samples=50)
    assert_(isinstance(m.get_samples(), dict))


def test_sample_kernel_custom_lscale_prior():
    lscale_prior_dist = numpyro.distributions.Normal(20, .1)
    m1 = MultiTaskGP(1, 'RBF', num_latents=2, num_tasks=2, rank=2)
    with numpyro.handlers.seed(rng_seed=1):
        lscale1 = m1._sample_kernel_params()["k_length"]
    m2 = MultiTaskGP(1, 'RBF', num_latents=2, num_tasks=2, rank=2, 
                     lengthscale_prior_dist=lscale_prior_dist)
    with numpyro.handlers.seed(rng_seed=1):
        lscale2 = m2._sample_kernel_params()["k_length"]
    assert_(not onp.array_equal(lscale1, lscale2))


def test_sample_task_kernel_custom_W_prior():
    W_prior_dist = numpyro.distributions.Normal(20*jnp.ones((2, 2, 2)), 0.1*jnp.ones((2, 2, 2)))
    m1 = MultiTaskGP(1, 'RBF', num_latents=2, num_tasks=2, rank=2)
    with numpyro.handlers.seed(rng_seed=1):
        W1 = m1._sample_task_kernel_params()["W"]
    m2 = MultiTaskGP(1, 'RBF', num_latents=2, num_tasks=2, rank=2,
                     W_prior_dist=W_prior_dist)
    with numpyro.handlers.seed(rng_seed=1):
        W2 = m2._sample_task_kernel_params()["W"]
    assert_(not onp.array_equal(W1, W2))


def test_sample_task_kernel_custom_v_prior():
    v_prior_dist = numpyro.distributions.Normal(20*jnp.ones((2, 2)), 0.1*jnp.ones((2, 2)))
    m1 = MultiTaskGP(1, 'RBF', num_latents=2, num_tasks=2, rank=2)
    with numpyro.handlers.seed(rng_seed=1):
        v1 = m1._sample_task_kernel_params()["v"]
    m2 = MultiTaskGP(1, 'RBF', num_latents=2, num_tasks=2, rank=2,
                     v_prior_dist=v_prior_dist)
    with numpyro.handlers.seed(rng_seed=1):
        v2 = m2._sample_task_kernel_params()["v"]
    assert_(not onp.array_equal(v1, v2))



"""
vgp.py
=======

Fully Bayesian implementation of Gaussian process regression for vector-valued functions

Created by Maxim Ziatdinov (email: maxim.ziatdinov@ai4microscopy.com)
"""

from functools import partial
from typing import Callable, Dict, Optional, Tuple, Union, Type

import jax
import jaxlib
import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
from jax import jit

from .gp import ExactGP


class vExactGP(ExactGP):
    """
    Gaussian process class for vector-valued targets

    Args:
        input_dim: number of input dimensions
        kernel: type of kernel ('RBF', 'Matern', 'Periodic')
        mean_fn: optional deterministic mean function (use 'mean_fn_priors' to make it probabilistic)
        kernel_prior: optional custom priors over kernel hyperparameters (uses LogNormal(0,1) by default)
        mean_fn_prior: optional priors over mean function parameters
        noise_prior: optional custom prior for observation noise
        noise_prior_dist:
            Optional custom prior distribution over observational noise. Defaults to LogNormal(0,1).
        lengthscale_prior_dist:
            Optional custom prior distribution over kernel lengthscale. Defaults to LogNormal(0, 1).

    """

    def __init__(self, input_dim: int, kernel: str,
                 mean_fn: Optional[Callable[[jnp.ndarray, Dict[str, jnp.ndarray]], jnp.ndarray]] = None,
                 kernel_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 mean_fn_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 noise_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 noise_prior_dist: Optional[dist.Distribution] = None,
                 lengthscale_prior_dist: Optional[dist.Distribution] = None
                 ) -> None:
        args = (input_dim, kernel, mean_fn,  kernel_prior, mean_fn_prior, noise_prior)
        super(vExactGP, self).__init__(*args)
        self.noise_prior_dist = noise_prior_dist
        self.lengthscale_prior_dist = lengthscale_prior_dist

    def model(self,
              X: jnp.ndarray,
              y: jnp.ndarray = None,
              **kwargs: float
              ) -> None:
        """GP probabilistic model with inputs X and vector-valued targets y"""
        jitter = kwargs.get("jitter", 1e-6)
        task_dim = X.shape[0]
        # Initialize mean function at zeros
        f_loc = jnp.zeros(X.shape[:2])
        # Sample parameters of kernels
        if self.kernel_prior:
            kernel_params = self.kernel_prior()
        else:
            kernel_params = self._sample_kernel_params(task_dim=task_dim)
        # Sample noise for each task
        if self.noise_prior:  # this will be removed in the future releases
            noise = self.noise_prior()
        else:
            noise = self._sample_noise(task_dim)
        # Add mean function (if any)
        if self.mean_fn is not None:
            args = [X]
            if self.mean_fn_prior is not None:
                args += [self.mean_fn_prior()]
            f_loc += self.mean_fn(*args).squeeze()
        # Compute kernels for each task in parallel
        jitter = jnp.array(jitter).repeat(task_dim)
        k_args = (X, X, kernel_params, noise)
        k = jax.vmap(self.kernel)(*k_args, jitter=jitter)
        # Sample y according to the standard Gaussian process formula
        numpyro.sample(
            "y",
            dist.MultivariateNormal(loc=f_loc, covariance_matrix=k),
            obs=y,
        )

    def _sample_noise(self, task_dim) -> jnp.ndarray:
        if self.noise_prior_dist is not None:
            noise_dist = self.noise_prior_dist
        else:
            noise_dist = dist.LogNormal(0, 1)
        with numpyro.plate("noise_plate", task_dim):
            noise = numpyro.sample("noise", noise_dist)
        return noise

    def _sample_kernel_params(self, task_dim: int = None) -> Dict[str, jnp.ndarray]:
        """
        Sample kernel parameters with default
        weakly-informative log-normal priors
        """
        if self.lengthscale_prior_dist is not None:
            length_dist = self.lengthscale_prior_dist
        else:
            length_dist = dist.LogNormal(0.0, 1.0)
        with numpyro.plate("plate_1", task_dim, dim=-2):  # task dimension
            with numpyro.plate('lengthscale', self.kernel_dim, dim=-1):  # allows using ARD kernel for kernel_dim > 1
                length = numpyro.sample("k_length", dist.LogNormal(0.0, 1.0))
        with numpyro.plate("plate_2", task_dim):  # task dimension'
            scale = numpyro.sample("k_scale", length_dist)
            if self.kernel_name == 'Periodic':
                period = numpyro.sample("period", dist.LogNormal(0.0, 1.0))
        kernel_params = {
            "k_length": length, "k_scale": scale,
            "period": period if self.kernel_name == "Periodic" else None}
        return kernel_params

    #@partial(jit, static_argnames='self')
    def _get_mvn_posterior(self,
                           X_train: jnp.ndarray, y_train: jnp.ndarray,
                           X_new: jnp.ndarray, params: Dict[str, jnp.ndarray],
                           m_X: Optional[jnp.ndarray] = None,
                           m_p: Optional[jnp.ndarray] = None,
                           noiseless: bool = False, **kwargs: float
                           ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        noise = params["noise"]
        noise_p = noise * (1 - jnp.array(noiseless, int))
        y_residual = y_train
        if m_X is not None:
            y_residual -= m_X
        # compute kernel matrices for train and test data
        k_pp = self.kernel(X_new, X_new, params, noise_p, **kwargs)
        k_pX = self.kernel(X_new, X_train, params, jitter=0.0)
        k_XX = self.kernel(X_train, X_train, params, noise, **kwargs)
        # compute the predictive covariance and mean
        K_xx_inv = jnp.linalg.inv(k_XX)
        cov = k_pp - jnp.matmul(k_pX, jnp.matmul(K_xx_inv, jnp.transpose(k_pX)))
        mean = jnp.matmul(k_pX, jnp.matmul(K_xx_inv, y_residual))
        if m_p is not None:
            mean += m_p
        return mean, cov

    def get_mvn_posterior(self,
                          X_new: jnp.ndarray, params: Dict[str, jnp.ndarray],
                          noiseless: bool = False, **kwargs: float
                          ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Returns parameters (mean and cov) of multivariate normal posterior
        for a single sample of GP parameters. Wrapper over self._get_mvn_posterior.
        """
        task_dim = X_new.shape[0]
        jitter = kwargs.get("jitter", 1e-6)
        if self.mean_fn is not None:  # Compute mean function for training data and new data
            get_args = lambda x: [x, params] if self.mean_fn_prior else [x]
            m_X = self.mean_fn(*get_args(self.X_train)).squeeze()
            m_p = self.mean_fn(*get_args(X_new)).squeeze()
            params_unsqueezed = {   # ensure all params has the same 'batch dimension' for vmap to work
                k: p[None].repeat(X_new.shape[0], axis=0) if p.ndim == 0 else p
                for k, p in params.items()
            }
            vmap_args = (
                self.X_train, self.y_train, X_new, params_unsqueezed, m_X, m_p)
        else:
            vmap_args = (self.X_train, self.y_train, X_new, params)
        noiseless = jnp.array(noiseless, int).repeat(task_dim)
        jitter = jnp.array(jitter).repeat(task_dim)
        mean, cov = jax.vmap(
            self._get_mvn_posterior)(*vmap_args, noiseless=noiseless, jitter=jitter)
        return mean, cov

    def predict_in_batches(self, rng_key: jnp.ndarray,
                           X_new: jnp.ndarray,  batch_size: int = 100,
                           samples: Optional[Dict[str, jnp.ndarray]] = None,
                           n: int = 1, filter_nans: bool = False,
                           predict_fn: Callable[[jnp.ndarray, int], Tuple[jnp.ndarray]] = None,
                           noiseless: bool = False,
                           device: Type[jaxlib.xla_extension.Device] = None,
                           **kwargs: float
                           ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Make prediction at X_new with sampled GP parameters
        by spitting the input array into chunks ("batches") and running
        predict_fn (defaults to self.predict) on each of them one-by-one
        to avoid a memory overflow
        """
        X_new = self._set_data(X_new)
        y_pred, y_sampled = self._predict_in_batches(
            rng_key, X_new, batch_size, 1, samples, n,
            filter_nans, predict_fn, noiseless, device, **kwargs)
        y_pred = jnp.concatenate(y_pred, -1)
        y_sampled = jnp.concatenate(y_sampled, -1)
        return y_pred, y_sampled

    def _set_data(self,
                  X: jnp.ndarray,
                  y: Optional[jnp.ndarray] = None
                  ) -> Union[Tuple[jnp.ndarray], jnp.ndarray]:
        X = X[..., None] if X.ndim == 2 else X  # add feature pseudo-dimension
        if y is not None:
            if y.shape[0] != X.shape[0]:
                raise AssertionError(
                    "Task dimensions must be identical in inputs and targets")
            return X, y
        return X



import sys
import numpy as onp
import jax.numpy as jnp
import jax.random as jra
from numpy.testing import assert_equal, assert_, assert_array_equal

sys.path.insert(0, "../gpax/")

from gpax.utils import preprocess_sparse_image, split_dict, random_sample_dict, get_keys


def test_sparse_img_processing():
    img = onp.random.randn(16, 16)
    # Generate random indices
    idx = [onp.random.randint(0, 16) for _ in range(100)], [onp.random.randint(0, 16) for _ in range(100)]
    # Set these indices to zero
    img[idx] = 0
    # Test the utility function
    X, y, X_full = preprocess_sparse_image(img)
    assert_equal(X.ndim, 2)
    assert_equal(y.ndim,  1)
    assert_equal(X_full.ndim, 2)
    assert_(X.shape[0] < 16*16)
    assert_equal(X.shape[1], 2)
    assert_equal(y.shape[0], X.shape[0])
    assert_equal(X_full.shape[0], 16*16)
    assert_equal(X_full.shape[1], 2)


def test_split_dict():
    data = {
        'a': jnp.array([1, 2, 3, 4, 5, 6]),
        'b': jnp.array([10, 20, 30, 40, 50, 60])
    }
    chunk_size = 4

    result = split_dict(data, chunk_size)

    expected = [
        {'a': jnp.array([1, 2, 3, 4]), 'b': jnp.array([10, 20, 30, 40])},
        {'a': jnp.array([5, 6]), 'b': jnp.array([50, 60])},
    ]

    # Check that the length of the result matches the expected length
    assert len(result) == len(expected)

    # Check that each chunk matches the expected chunk
    for r, e in zip(result, expected):
        for k in data:
            assert_array_equal(r[k], e[k])


def test_random_sample_size():
    data = {
        'a': jnp.array([1, 2, 3, 4, 5]),
        'b': jnp.array([5, 4, 3, 2, 1]),
        'c': jnp.array([10, 20, 30, 40, 50])
    }
    num_samples = 3
    rng_key = jra.PRNGKey(123)
    sampled_data = random_sample_dict(data, num_samples, rng_key)
    for value in sampled_data.values():
        assert_(len(value) == num_samples)


def test_random_sample_consistency():
    data = {
        'a': jnp.array([1, 2, 3, 4, 5]),
        'b': jnp.array([5, 4, 3, 2, 1]),
        'c': jnp.array([10, 20, 30, 40, 50])
    }
    num_samples = 3
    rng_key = jra.PRNGKey(123)
    sampled_data1 = random_sample_dict(data, num_samples, rng_key)
    sampled_data2 = random_sample_dict(data, num_samples, rng_key)

    for key in sampled_data1:
        assert_(jnp.array_equal(sampled_data1[key], sampled_data2[key]))


def test_random_sample_difference():
    data = {
        'a': jnp.array([1, 2, 3, 4, 5]),
        'b': jnp.array([5, 4, 3, 2, 1]),
        'c': jnp.array([10, 20, 30, 40, 50])
    }
    num_samples = 3
    rng_key1, rng_key2 = get_keys()
    sampled_data1 = random_sample_dict(data, num_samples, rng_key1)
    sampled_data2 = random_sample_dict(data, num_samples, rng_key2)

    for key in sampled_data1:
        assert_(not jnp.array_equal(sampled_data1[key], sampled_data2[key]))



import sys
import pytest
import numpy as onp
import jax
import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
from numpy.testing import assert_equal, assert_

sys.path.insert(0, "../gpax/")

from gpax.models.gp import ExactGP
from gpax.models.vidkl import viDKL
from gpax.utils import get_keys
from gpax.acquisition.base_acq import ei, ucb, poi, ue, kg
from gpax.acquisition.acquisition import _compute_mean_and_var
from gpax.acquisition.acquisition import EI, UCB, UE, POI, Thompson, KG
from gpax.acquisition.batch_acquisition import _compute_batch_acquisition
from gpax.acquisition.batch_acquisition import qEI, qPOI, qUCB, qKG
from gpax.acquisition.penalties import compute_penalty


class mock_GP:
    def __init__(self):
        self.mcmc = 1

    def get_samples(self):
        rng_key = get_keys()[1]
        samples = {"k_length": jax.random.normal(rng_key, shape=(100, 1)),
                   "k_scale": jax.random.normal(rng_key, shape=(100,)),
                   "noise": jax.random.normal(rng_key, shape=(100,))}
        return samples


@pytest.mark.parametrize("base_acq", [ei, ucb, poi, ue])
def test_base_standard_acq(base_acq):
    mean = onp.random.randn(10,)
    var = onp.random.uniform(0, 1, size=10)
    moments = (mean, var)
    obj = base_acq(moments)
    assert_(isinstance(obj, jnp.ndarray))
    assert_equal(len(obj), len(mean))
    assert_equal(obj.ndim, 1)


def test_base_acq_kg():
    rng_keys = get_keys()
    X = onp.random.randn(8,)
    X_new = onp.random.randn(12, 1)
    y = 10 * X**2
    m = ExactGP(1, 'RBF')
    m.fit(rng_keys[0], X, y, num_warmup=100, num_samples=100)
    sample = {k: v[0] for (k, v) in m.get_samples().items()}
    obj = kg(m, X_new, sample)
    assert_(isinstance(obj, jnp.ndarray))
    assert_equal(len(obj), len(X_new))
    assert_equal(obj.ndim, 1)


@pytest.mark.parametrize("base_acq", [ei, ucb, poi])
def test_base_standard_acq_maximize(base_acq):
    mean = onp.random.randn(10,)
    var = onp.random.uniform(0, 1, size=10)
    moments = (mean, var)
    obj1 = base_acq(moments, maximize=False)
    obj2 = base_acq(moments, maximize=True)
    assert_(not onp.array_equal(obj1, obj2))


@pytest.mark.parametrize("base_acq", [ei, poi])
def test_base_standard_acq_best_f(base_acq):
    mean = onp.random.randn(10,)
    var = onp.random.uniform(0, 1, size=10)
    best_f = mean.min() - 0.01
    moments = (mean, var)
    obj1 = base_acq(moments)
    obj2 = base_acq(moments, best_f=best_f)
    assert_(not onp.array_equal(obj1, obj2))


def test_compute_mean_and_var():
    rng_keys = get_keys()
    X = onp.random.randn(8,)
    X_new = onp.random.randn(12,)
    y = 10 * X**2
    m = ExactGP(1, 'RBF')
    m.fit(rng_keys[0], X, y, num_warmup=100, num_samples=100)
    mean, var = _compute_mean_and_var(
        rng_keys[1], m, X_new, n=1, noiseless=True)
    assert_equal(mean.shape, (len(X_new),))
    assert_equal(var.shape, (len(X_new),))


@pytest.mark.parametrize("acq", [EI, UCB, UE, Thompson, POI])
def test_acq_gp(acq):
    rng_keys = get_keys()
    X = onp.random.randn(8,)
    X_new = onp.random.randn(12,)
    y = 10 * X**2
    m = ExactGP(1, 'RBF')
    m.fit(rng_keys[0], X, y, num_warmup=100, num_samples=100)
    obj = acq(rng_keys[1], m, X_new)
    assert_(isinstance(obj, jnp.ndarray))
    assert_equal(obj.squeeze().shape, (len(X_new),))


@pytest.mark.parametrize("acq", [EI, UCB, UE, Thompson, POI, KG])
def test_acq_dkl(acq):
    rng_keys = get_keys()
    X = onp.random.randn(8, 10)
    X_new = onp.random.randn(12, 10)
    y = (10 * X**2).mean(-1)
    m = viDKL(1, 2, 'RBF')
    m.fit(rng_keys[0], X, y, num_steps=10)
    obj = acq(rng_keys[1], m, X_new)
    assert_(isinstance(obj, jnp.ndarray))
    assert_equal(obj.shape, (len(X_new),))


def test_UCB_beta():
    rng_keys = get_keys()
    X = onp.random.randn(8,)
    X_new = onp.random.randn(12,)
    y = 10 * X**2
    m = ExactGP(1, 'RBF')
    m.fit(rng_keys[0], X, y, num_warmup=100, num_samples=100)
    obj1 = UCB(rng_keys[1], m, X_new, beta=2)
    obj2 = UCB(rng_keys[1], m, X_new, beta=4)
    obj3 = UCB(rng_keys[1], m, X_new, beta=2)
    assert_(not onp.array_equal(obj1, obj2))
    assert_(onp.array_equal(obj1, obj3))


def test_EI_gp_penalty_inv_distance():
    rng_keys = get_keys()
    X = onp.random.randn(8,)
    X_new = onp.random.randn(12,)
    recent_points = X_new[-3:-1]
    y = 10 * X**2
    m = ExactGP(1, 'RBF')
    m.fit(rng_keys[0], X, y, num_warmup=100, num_samples=100)
    obj1 = EI(rng_keys[1], m, X_new)
    obj2 = EI(rng_keys[1], m, X_new, penalty="inverse_distance", recent_points=recent_points)
    assert_(obj2[-1] < obj1[-1])
    assert_(obj2[-2] < obj1[-2])


def test_UCB_gp_penalty_inv_distance():
    rng_keys = get_keys()
    X = onp.random.randn(8,)
    X_new = onp.random.randn(12,)
    recent_points = X_new[-3:-1]
    y = 10 * X**2
    m = ExactGP(1, 'RBF')
    m.fit(rng_keys[0], X, y, num_warmup=100, num_samples=100)
    obj1 = UCB(rng_keys[1], m, X_new)
    obj2 = UCB(rng_keys[1], m, X_new, penalty="inverse_distance", recent_points=recent_points)
    assert_(obj2[-1] < obj1[-1])
    assert_(obj2[-2] < obj1[-2])


def test_UE_gp_penalty_inv_distance():
    rng_keys = get_keys()
    X = onp.random.randn(8,)
    X_new = onp.random.randn(12,)
    recent_points = X_new[-3:-1]
    y = 10 * X**2
    m = ExactGP(1, 'RBF')
    m.fit(rng_keys[0], X, y, num_warmup=100, num_samples=100)
    obj1 = UE(rng_keys[1], m, X_new)
    obj2 = UE(rng_keys[1], m, X_new, penalty="inverse_distance", recent_points=recent_points)
    assert_(obj2[-1] < obj1[-1])
    assert_(obj2[-2] < obj1[-2])


@pytest.mark.parametrize("acq", [EI, UCB, UE, Thompson])
def test_acq_dkl(acq):
    rng_keys = get_keys()
    X = onp.random.randn(32, 36)
    y = onp.random.randn(32,)
    X_new = onp.random.randn(10, 36)
    m = viDKL(X.shape[-1])
    m.fit(rng_keys[0], X, y, num_steps=20, step_size=0.05)
    obj = acq(rng_keys[1], m, X_new)
    assert_(isinstance(obj, jnp.ndarray))
    assert_equal(obj.squeeze().shape, (len(X_new),))


@pytest.mark.parametrize("maximize_distance", [False, True])
def test_compute_batch_acquisition(maximize_distance):
    def mock_acq_fn(*args):
        return jnp.arange(0, 10)
    X = onp.random.randn(10)
    rng_key = get_keys()[0]
    m = mock_GP()
    obj = _compute_batch_acquisition(
        rng_key, m, X, mock_acq_fn, subsample_size=7,
        maximize_distance=maximize_distance)
    assert_equal(obj.shape[0], 7)


@pytest.mark.parametrize("q", [1, 3])
@pytest.mark.parametrize("acq", [qEI, qPOI, qUCB, qKG])
def test_batched_acq(acq, q):
    rng_key = get_keys()
    X = onp.random.randn(8,)
    X_new = onp.random.randn(12,)
    y = 10 * X**2
    m = ExactGP(1, 'RBF')
    m.fit(rng_key[0], X, y, num_warmup=100, num_samples=100)
    obj = acq(rng_key[1], m, X_new, subsample_size=q)
    assert_equal(obj.shape, (q, len(X_new)))


@pytest.mark.parametrize('pen', ['delta', 'inverse_distance'])
@pytest.mark.parametrize("acq", [EI, UCB, UE])
def test_acq_penalty_indices(acq, pen):
    rng_keys = get_keys()
    h = w = 5
    X = onp.random.randn(h*w, 16)
    y = onp.random.randn(len(X))
    indices = onp.array([(i, j) for i in range(h) for j in range(w)])
    X_new = onp.random.randn(h*w, 16)
    m = viDKL(input_dim=16)
    m.fit(rng_keys[0], X, y, num_steps=50)
    obj1 = acq(rng_keys[1], m, X_new, grid_indices=indices,
               penalty=pen, recent_points=indices[5:7])
    obj2 = acq(rng_keys[1], m, X_new)
    assert_(isinstance(obj1, jnp.ndarray))
    assert_equal(obj1.squeeze().shape, (len(X_new),))
    assert_(onp.count_nonzero(obj1 - obj2) > 0)


def test_compute_penalty_delta():
    X = jnp.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    recent_points = jnp.array([[4, 5, 6], [1, 2, 3]])
    penalty_factor = 1.0
    penalties = compute_penalty(X, recent_points, "delta", penalty_factor)
    assert jnp.array_equal(penalties, jnp.array([jnp.inf, jnp.inf, 0.0]))


def test_compute_penalty_inverse_distance():
    X = jnp.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    recent_points = jnp.array([[4, 5, 6], [7, 8, 9]])
    penalty_factor = 1.0
    penalties = compute_penalty(X, recent_points, "inverse_distance", penalty_factor)
    assert_(penalties[-1] > penalties[-2])
    assert_(penalties[-2] > penalties[-3])


@pytest.mark.parametrize("acq_func", [EI, UCB, UE])
def test_acq_error(acq_func):
    # Initialize the required inputs
    rng_keys = get_keys()
    X = onp.random.randn(8,)
    X_new = onp.random.randn(12,)
    recent_points = X_new[-3:-1]
    y = 10 * X**2
    m = ExactGP(1, 'RBF')
    m.fit(rng_keys[0], X, y, num_warmup=100, num_samples=100)

    # Test the ValueError for not passing the recent_points when penalty is not None
    with pytest.raises(ValueError):
        acq_func(rng_keys[1], m, X_new, penalty='delta')

    # Test the ValueError for passing recent_points as non ndarray type
    with pytest.raises(ValueError):
        acq_func(rng_keys[1], m, X_new, penalty='delta', recent_points=[1, 2, 3])

    # Test the function with correct parameters
    try:
        output = acq_func(rng_keys[1], m, X, penalty='delta', recent_points=recent_points)
        assert isinstance(output, jnp.ndarray)  # check if output is of correct type
    except Exception as e:
        pytest.fail(f"Test failed, error: {str(e)}")



from .kernels import (MaternKernel, NNGPKernel, PeriodicKernel, RBFKernel,
                      get_kernel, nngp_erf, nngp_relu)
from .mtkernels import (LCMKernel, MultitaskKernel, MultivariateKernel,
                        index_kernel)

__all__ = [
    "RBFKernel",
    "MaternKernel",
    "PeriodicKernel",
    "NNGPKernel",
    "get_kernel",
    "index_kernel",
    "MultitaskKernel",
    "MultivariateKernel",
    "LCMKernel"
]



"""
dkl.py
=======

Fully Bayesian implementation of deep kernel learning

Created by Maxim Ziatdinov (email: maxim.ziatdinov@ai4microscopy.com)
"""

from functools import partial
from typing import Callable, Dict, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
from jax import jit

from .vgp import vExactGP


class DKL(vExactGP):
    """
    Fully Bayesian implementation of deep kernel learning

    Args:
        input_dim:
            Number of input dimensions
        z_dim:
            Latent space dimensionality (defaults to 2)
        kernel:
            Kernel function ('RBF', 'Matern', 'Periodic', or custom function)
        kernel_prior:
            Optional priors over kernel hyperparameters; uses LogNormal(0,1) by default
        nn:
            Custom neural network ('feature extractor'); uses a 3-layer MLP
            with hyperbolic tangent activations by default
        nn_prior:
            Priors over the weights and biases in 'nn'; uses normal priors by default
        latent_prior:
            Optional prior over the latent space (BNN embedding); uses none by default

        **kwargs:
            Optional custom prior distributions over observational noise (noise_dist_prior)
            and kernel lengthscale (lengthscale_prior_dist)


    Examples:

        DKL with image patches as inputs and a 1-d vector as targets

        >>> # Get random number generator keys for training and prediction
        >>> key1, key2 = gpax.utils.get_keys()
        >>> input data dimensions are (n, height*width*channels)
        >>> data_dim = X.shape[-1]
        >>> # Initialize DKL model with 2 latent dimensions
        >>> dkl = gpax.DKL(data_dim, z_dim=2, kernel='RBF')
        >>> # Train model by parallelizing HMC chains on a single GPU
        >>> dkl.fit(key1, X, y, num_warmup=333, num_samples=333, num_chains=3, chain_method='vectorized')
        >>> # Obtain posterior mean and samples from DKL posterior at new inputs
        >>> # using batches to avoid memory overflow
        >>> y_pred, y_samples = dkl.predict_in_batches(key2, X_new)
    """

    def __init__(self, input_dim: int, z_dim: int = 2, kernel: str = 'RBF',
                 kernel_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 nn: Optional[Callable[[jnp.ndarray, Dict[str, jnp.ndarray]], jnp.ndarray]] = None,
                 nn_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 latent_prior: Optional[Callable[[jnp.ndarray], Dict[str, jnp.ndarray]]] = None,
                 **kwargs
                 ) -> None:
        super(DKL, self).__init__(input_dim, kernel, None, kernel_prior, **kwargs)
        self.nn = nn if nn else mlp
        self.nn_prior = nn_prior if nn_prior else mlp_prior(input_dim, z_dim)
        self.kernel_dim = z_dim
        self.latent_prior = latent_prior

    def model(self,
              X: jnp.ndarray,
              y: jnp.ndarray = None,
              **kwargs: float
              ) -> None:
        """DKL probabilistic model"""
        jitter = kwargs.get("jitter", 1e-6)
        task_dim = X.shape[0]
        # BNN part
        bnn_params = self.nn_prior(task_dim)
        z = jax.jit(jax.vmap(self.nn))(X, bnn_params)
        if self.latent_prior:  # Sample latent variable
            z = self.latent_prior(z)
        # Sample GP kernel parameters
        if self.kernel_prior:
            kernel_params = self.kernel_prior()
        else:
            kernel_params = self._sample_kernel_params(task_dim)
        # Sample noise
        noise = self._sample_noise(task_dim)
        # GP's mean function
        f_loc = jnp.zeros(z.shape[:2])
        # compute kernel(s)
        jitter = jnp.array(jitter).repeat(task_dim)
        k_args = (z, z, kernel_params, noise)
        k = jax.vmap(self.kernel)(*k_args, jitter=jitter)
        # Sample y according to the standard Gaussian process formula
        numpyro.sample(
            "y",
            dist.MultivariateNormal(loc=f_loc, covariance_matrix=k),
            obs=y,
        )

    #@partial(jit, static_argnames='self')
    def _get_mvn_posterior(self,
                           X_train: jnp.ndarray, y_train: jnp.ndarray,
                           X_new: jnp.ndarray, params: Dict[str, jnp.ndarray],
                           noiseless: bool = False, **kwargs: float
                           ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        noise = params["noise"]
        noise_p = noise * (1 - jnp.array(noiseless, int))
        # embed data into the latent space
        z_train = self.nn(X_train, params)
        z_new = self.nn(X_new, params)
        # compute kernel matrices for train and new ('test') data
        k_pp = self.kernel(z_new, z_new, params, noise_p, **kwargs)
        k_pX = self.kernel(z_new, z_train, params, jitter=0.0)
        k_XX = self.kernel(z_train, z_train, params, noise, **kwargs)
        # compute the predictive covariance and mean
        K_xx_inv = jnp.linalg.inv(k_XX)
        cov = k_pp - jnp.matmul(k_pX, jnp.matmul(K_xx_inv, jnp.transpose(k_pX)))
        mean = jnp.matmul(k_pX, jnp.matmul(K_xx_inv, y_train))
        return mean, cov

    @partial(jit, static_argnames='self')
    def embed(self, X_new: jnp.ndarray) -> jnp.ndarray:
        """
        Embeds data into the latent space using the inferred weights
        of the DKL's Bayesian neural network
        """
        samples = self.get_samples(chain_dim=False)
        predictive = jax.vmap(lambda params: self.nn(X_new, params))
        z = predictive(samples)
        return z

    def _set_data(self,
                  X: jnp.ndarray,
                  y: Optional[jnp.ndarray] = None
                  ) -> Union[Tuple[jnp.ndarray], jnp.ndarray]:
        X = X[None] if X.ndim == 2 else X  # add task pseudo-dimension
        if y is not None:
            y = y[None] if y.ndim == 1 else y  # add task pseudo-dimension
            return X, y
        return X

    def _print_summary(self):
        list_of_keys = ["k_scale", "k_length", "noise", "period"]
        samples = self.get_samples(1)
        numpyro.diagnostics.print_summary(
            {k: v for (k, v) in samples.items() if k in list_of_keys})


def sample_weights(name: str, in_channels: int, out_channels: int, task_dim: int) -> jnp.ndarray:
    """Sampling weights matrix"""
    with numpyro.plate("batch_dim", task_dim, dim=-3):
        w = numpyro.sample(name=name, fn=dist.Normal(
            loc=jnp.zeros((in_channels, out_channels)),
            scale=jnp.ones((in_channels, out_channels))))
    return w


def sample_biases(name: str, channels: int, task_dim: int) -> jnp.ndarray:
    """Sampling bias vector"""
    with numpyro.plate("batch_dim", task_dim, dim=-3):
        b = numpyro.sample(name=name, fn=dist.Normal(
            loc=jnp.zeros((channels)), scale=jnp.ones((channels))))
    return b


def mlp(X: jnp.ndarray, params: Dict[str, jnp.ndarray]) -> jnp.ndarray:
    """Simple MLP for a single MCMC sample of weights and biases"""
    h1 = jnp.tanh(jnp.matmul(X, params["w1"]) + params["b1"])
    h2 = jnp.tanh(jnp.matmul(h1, params["w2"]) + params["b2"])
    z = jnp.matmul(h2, params["w3"]) + params["b3"]
    return z


def mlp_prior(input_dim: int, zdim: int = 2) -> Dict[str, jnp.array]:
    """Priors over weights and biases in the default Bayesian MLP"""
    hdim = [64, 32]

    def _bnn_prior(task_dim: int):
        w1 = sample_weights("w1", input_dim, hdim[0], task_dim)
        b1 = sample_biases("b1", hdim[0], task_dim)
        w2 = sample_weights("w2", hdim[0], hdim[1], task_dim)
        b2 = sample_biases("b2", hdim[1], task_dim)
        w3 = sample_weights("w3", hdim[1], zdim, task_dim)
        b3 = sample_biases("b3", zdim, task_dim)
        return {"w1": w1, "b1": b1, "w2": w2, "b2": b2, "w3": w3, "b3": b3}

    return _bnn_prior



"""
spm.py
=======

Bayesian inference with structured probabilistic model. Serves as a wrapper
over the NumPyro's functions for Bayesian inference on probabilistic models.
While it has no direct connection to Gaussian processes, this module may come
in handy for comparisons between predicitons of GP and classical Bayesian models.

Created by Maxim Ziatdinov (email: maxim.ziatdinov@ai4microscopy.com)
"""

from typing import Callable, Optional, Tuple, Type, Dict

import jax
import jaxlib
import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
from numpyro.infer import MCMC, NUTS, Predictive, init_to_median

model_type = Callable[[jnp.ndarray, Dict[str, jnp.ndarray]], jnp.ndarray]
prior_type = Callable[[], Dict[str, jnp.ndarray]]


class sPM:
    """
    Bayesian inference with structured probabilistic model.
    Serves as a wrapper over the NumPyro's functions for Bayesian inference
    on probabilistic models.

    Args:
        model:
            Deterministic model of expected system's behavior.
        model_prior:
            Priors over model parameters
        noise_prior:
            Optional custom prior for observation noise;
            uses LogNormal(0,1) by default.
    """
    def __init__(self,
                 model: model_type,
                 model_prior: prior_type,
                 noise_prior: Optional[prior_type] = None) -> None:
        self._model = model
        self.model_prior = model_prior
        if noise_prior is None:
            self.noise_prior = lambda: numpyro.sample("sig", dist.LogNormal(0, 1))
        else:
            self.noise_prior = noise_prior
        self.mcmc = None

    def model(self, X: jnp.ndarray, y: jnp.ndarray = None) -> None:
        """
        Full probabilistic model
        """
        # Sample model parameters
        params = self.model_prior()
        # Compute the function's value
        mu = numpyro.deterministic("mu", self._model(X, params))
        # Sample observational noise
        sig = self.noise_prior()
        # Score against the observed data points
        numpyro.sample("y", dist.Normal(mu, sig), obs=y)

    def fit(self, rng_key: jnp.array, X: jnp.ndarray, y: jnp.ndarray,
            num_warmup: int = 2000, num_samples: int = 2000,
            num_chains: int = 1, chain_method: str = 'sequential',
            progress_bar: bool = True, print_summary: bool = True,
            device: Type[jaxlib.xla_extension.Device] = None) -> None:
        """
        Run HMC to infer parameters of the structured probabilistic model

        Args:
            rng_key: random number generator key
            X: 1D or 2D 'feature vector' with :math:`(n,)` or :math:`n x num_features` dimensions
            y: 1D 'target vector' with :math:`(n,)` dimensions
            num_warmup: number of HMC warmup states
            num_samples: number of HMC samples
            num_chains: number of HMC chains
            chain_method: 'sequential', 'parallel' or 'vectorized'
            progress_bar: show progress bar
            print_summary: print summary at the end of sampling
            device:
                optionally specify a cpu or gpu device on which to run the inference;
                e.g., ``device=jax.devices("cpu")[0]`` 
        """
        if device:
            X = jax.device_put(X, device)
            y = jax.device_put(y, device)
        init_strategy = init_to_median(num_samples=10)
        kernel = NUTS(self.model, init_strategy=init_strategy)
        self.mcmc = MCMC(
            kernel,
            num_warmup=num_warmup,
            num_samples=num_samples,
            num_chains=num_chains,
            chain_method=chain_method,
            progress_bar=progress_bar,
            jit_model_args=False
        )
        self.mcmc.run(rng_key, X, y)
        if print_summary:
            self._print_summary()

    def get_samples(self, chain_dim: bool = False) -> Dict[str, jnp.ndarray]:
        """Get posterior samples (after running the MCMC chains)"""
        return self.mcmc.get_samples(group_by_chain=chain_dim)

    def get_param_means(self):
        """
        Returns mean value for each probabilistic parameter in the model
        """
        samples = self.get_samples()
        param_means = {
            k: v.mean(0).item() for (k, v) in samples.items() if k != 'mu'
        }
        return param_means

    def sample_from_prior(self, rng_key: jnp.ndarray,
                          X: jnp.ndarray, num_samples: int = 10):
        """
        Samples from prior predictive distribution at X
        """
        prior_predictive = Predictive(self.model, num_samples=num_samples)
        samples = prior_predictive(rng_key, X)
        return samples['y']

    def predict(self, rng_key: jnp.ndarray, X_new: jnp.ndarray,
                samples: Optional[Dict[str, jnp.ndarray]] = None,
                filter_nans: bool = False, take_point_predictions_mean: bool = True,
                device: Type[jaxlib.xla_extension.Device] = None
                ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Make prediction at X_new points using sampled GP hyperparameters

        Args:
            rng_key: random number generator key
            X_new: 2D vector with new/'test' data of :math:`n x num_features` dimensionality
            samples: optional posterior samples
            filter_nans: filter out samples containing NaN values (if any)
            take_point_predictions_mean: take a mean of point predictions (without sampling from the normal distribution)
            device:
                optionally specify a cpu or gpu device on which to make a prediction;
                e.g., ```device=jax.devices("gpu")[0]```

        Returns:
            Point predictions (or their mean) and posterior predictive distribution
        """
        if samples is None:
            samples = self.get_samples(chain_dim=False)
        if device:
            X_new = jax.device_put(X_new, device)
            samples = jax.device_put(samples, device)
        predictive = Predictive(
            self.model, posterior_samples=samples, parallel=True)
        y_pred = predictive(rng_key, X_new)
        y_pred, y_sampled = y_pred["mu"], y_pred["y"]
        if filter_nans:
            y_sampled_ = [y_i for y_i in y_sampled if not jnp.isnan(y_i).any()]
            y_sampled = jnp.array(y_sampled_)
        if take_point_predictions_mean:
            y_pred = y_pred.mean(0)
        return y_pred, y_sampled

    def _print_summary(self):
        self.mcmc.print_summary()



"""
vi_ibnn.py
=======

Infinite width Bayesian neural net (variational approximation)

Created by Maxim Ziatdinov (email: maxim.ziatdinov@ai4microscopy.com)
"""

from typing import Optional, Dict, Callable

import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist

from .vigp import viGP
from ..kernels import get_kernel


class vi_iBNN(viGP):
    """
    Variational inference-based infinite Bayesian neural net (vi-iBNN)

    Args:
        input_dim:
            Number of input dimensions
        depth:
            The number of layers in the corresponding infinite-width neural network.
        activation:
            activation function ('erf' or 'relu')
        mean_fn:
            Optional deterministic mean function (use 'mean_fn_priors' to make it probabilistic)
        nngp_prior:
            Optional custom priors over NNGP kernel hyperparameters.
            The defaults are HalfNormal(1) for bias variance and LogNormal(0, 10) for weight variance.
        mean_fn_prior:
            Optional priors over mean function parameters
        noise_prior:
            Optional custom prior for observation noise; uses LogNormal(0,1) by default.
    """

    def __init__(self, input_dim: int, depth: int = 3, activation: str = 'erf',
                 mean_fn: Optional[Callable[[jnp.ndarray, Dict[str, jnp.ndarray]], jnp.ndarray]] = None,
                 nngp_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 mean_fn_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 noise_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None
                 ) -> None:
        args = (input_dim, None, mean_fn, nngp_prior, mean_fn_prior, noise_prior)
        super(vi_iBNN, self).__init__(*args)
        self.kernel = get_kernel("NNGP", activation=activation, depth=depth)

    def _sample_kernel_params(self) -> Dict[str, jnp.ndarray]:
        """
        Sample NNGP kernel parameters with default
        weakly-informative log-normal priors
        """
        var_b = numpyro.sample("var_b", dist.HalfNormal(1))
        var_w = numpyro.sample("var_w", dist.LogNormal(0, 10))
        return {"var_b": var_b, "var_w": var_w}



import sys
import pytest
import numpy as onp
import jax.numpy as jnp
from numpy.testing import assert_equal


sys.path.insert(0, "../gpax/")

from gpax.models import iBNN, vi_iBNN
from gpax.utils import get_keys


def get_dummy_data(jax_ndarray=True, unsqueeze=False):
    X = onp.linspace(1, 2, 8) + 0.1 * onp.random.randn(8,)
    y = (10 * X**2)
    if unsqueeze:
        X = X[:, None]
    if jax_ndarray:
        return jnp.array(X), jnp.array(y)
    return X, y


@pytest.mark.parametrize("activation", ['erf', 'relu'])
@pytest.mark.parametrize("depth", [1, 2, 3])
def test_ibnn_fit_predict(activation, depth):
    rng_keys = get_keys()
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    m = iBNN(1, depth=depth, activation=activation)
    m.fit(rng_keys[0], X, y, num_warmup=100, num_samples=100)
    y_pred, y_sampled = m.predict(rng_keys[1], X_test)
    assert isinstance(y_pred, jnp.ndarray)
    assert isinstance(y_sampled, jnp.ndarray)
    assert_equal(y_pred.shape, X_test.squeeze().shape)
    assert_equal(y_sampled.shape, (100, 1, X_test.shape[0]))


@pytest.mark.parametrize("activation", ['erf', 'relu'])
@pytest.mark.parametrize("depth", [1, 2, 3])
def test_viibnn_fit_predict(activation, depth):
    rng_keys = get_keys()
    X, y = get_dummy_data()
    X_test, _ = get_dummy_data()
    m = vi_iBNN(1, depth=depth, activation=activation)
    m.fit(rng_keys[0], X, y, num_steps=100)
    y_pred, y_var = m.predict(rng_keys[1], X_test)
    assert isinstance(y_pred, jnp.ndarray)
    assert isinstance(y_var, jnp.ndarray)
    assert_equal(y_pred.shape, y_var.shape)
    assert_equal(y_pred.shape, X_test.squeeze().shape)



"""
gp.py
=======

Fully Bayesian implementation of Gaussian process regression

Created by Maxim Ziatdinov (email: maxim.ziatdinov@ai4microscopy.com)
"""

import warnings
from functools import partial
from typing import Callable, Dict, Optional, Tuple, Type, Union

import jax
import jaxlib
import jax.numpy as jnp
import jax.random as jra
import numpyro
import numpyro.distributions as dist
from jax import jit
from numpyro.infer import MCMC, NUTS, init_to_median, Predictive

from ..kernels import get_kernel
from ..utils import split_in_batches

kernel_fn_type = Callable[[jnp.ndarray, jnp.ndarray, Dict[str, jnp.ndarray], jnp.ndarray],  jnp.ndarray]

clear_cache = jax._src.dispatch.xla_primitive_callable.cache_clear


class ExactGP:
    """
    Gaussian process class

    Args:
        input_dim:
            Number of input dimensions
        kernel:
            Kernel function ('RBF', 'Matern', 'Periodic', or custom function)
        mean_fn:
            Optional deterministic mean function (use 'mean_fn_priors' to make it probabilistic)
        kernel_prior:
            Optional custom priors over kernel hyperparameters. Use it when passing your custom kernel.
        mean_fn_prior:
            Optional priors over mean function parameters
        noise_prior_dist:
            Optional custom prior distribution over observational noise. Defaults to LogNormal(0,1).
        lengthscale_prior_dist:
            Optional custom prior distribution over kernel lengthscale. Defaults to LogNormal(0, 1).

    Examples:

        Regular GP for sparse noisy obervations

        >>> # Get random number generator keys for training and prediction
        >>> rng_key, rng_key_predict = gpax.utils.get_keys()
        >>> # Initialize model
        >>> gp_model = gpax.ExactGP(input_dim=1, kernel='Matern')
        >>> # Run HMC to obtain posterior samples for the GP model parameters
        >>> gp_model.fit(rng_key, X, y)  # X and y are arrays with dimensions (n, 1) and (n,)
        >>> # Make a noiseless prediction on new inputs
        >>> y_pred, y_samples = gp_model.predict(rng_key_predict, X_new, noiseless=True)

        GP with custom noise prior
        
        >>> gp_model = gpax.ExactGP(
        >>>     input_dim=1, kernel='RBF',
        >>>     noise_prior_dist = numpyro.distributions.HalfNormal(.1)
        >>> )
        >>> # Run HMC to obtain posterior samples for the GP model parameters
        >>> gp_model.fit(rng_key, X, y)  # X and y are arrays with dimensions (n, 1) and (n,)
        >>> # Make a noiselsess prediction on new inputs
        >>> y_pred, y_samples = gp_model.predict(rng_key_predict, X_new, noiseless=True)

        GP with custom probabilistic model as its mean function
        
        >>> # Define a deterministic mean function
        >>> mean_fn = lambda x, param: param["a"]*x + param["b"]
        >>>
        >>> # Define priors over the mean function parameters (to make it probabilistic)
        >>> def mean_fn_prior():
        >>>     a = numpyro.sample("a", numpyro.distributions.Normal(3, 1))
        >>>     b = numpyro.sample("b", numpyro.distributions.Normal(0, 1))
        >>>     return {"a": a, "b": b}
        >>>
        >>> # Initialize structural GP model
        >>> sgp_model = gpax.ExactGP(
                input_dim=1, kernel='Matern',
                mean_fn=mean_fn, mean_fn_prior=mean_fn_prior)
        >>> # Run HMC to obtain posterior samples for the GP model parameters
        >>> sgp_model.fit(rng_key, X, y)  # X and y are numpy arrays with dimensions (n, d) and (n,)
        >>> # Make a noiselsess prediction on new inputs
        >>> y_pred, y_samples = gp_model.predict(rng_key_predict, X_new, noiseless=True)
    """

    def __init__(self, input_dim: int, kernel: Union[str, kernel_fn_type],
                 mean_fn: Optional[Callable[[jnp.ndarray, Dict[str, jnp.ndarray]], jnp.ndarray]] = None,
                 kernel_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 mean_fn_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 noise_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 noise_prior_dist: Optional[dist.Distribution] = None,
                 lengthscale_prior_dist: Optional[dist.Distribution] = None
                 ) -> None:
        clear_cache()
        if noise_prior is not None:
            warnings.warn("`noise_prior` is deprecated and will be removed in a future version. "
                          "Please use `noise_prior_dist` instead, which accepts an instance of a "
                          "numpyro.distributions Distribution object, e.g., `dist.HalfNormal(scale=0.1)`, "
                          "rather than a function that calls `numpyro.sample`.", FutureWarning)   
        if kernel_prior is not None:
            warnings.warn("`kernel_prior` will remain available for complex priors. However, for "
                          "modifying only the lengthscales, it is recommended to use `lengthscale_prior_dist` instead. "
                          "`lengthscale_prior_dist` accepts an instance of a numpyro.distributions Distribution object, "
                          "e.g., `dist.Gamma(2, 5)`, rather than a function that calls `numpyro.sample`.", UserWarning)
        self.kernel_dim = input_dim
        self.kernel = get_kernel(kernel)
        self.kernel_name = kernel if isinstance(kernel, str) else None
        self.mean_fn = mean_fn
        self.kernel_prior = kernel_prior
        self.mean_fn_prior = mean_fn_prior
        self.noise_prior = noise_prior
        self.noise_prior_dist = noise_prior_dist
        self.lengthscale_prior_dist = lengthscale_prior_dist
        self.X_train = None
        self.y_train = None
        self.mcmc = None

    def model(self,
              X: jnp.ndarray,
              y: jnp.ndarray = None,
              **kwargs: float
              ) -> None:
        """GP probabilistic model with inputs X and targets y"""
        # Initialize mean function at zeros
        f_loc = jnp.zeros(X.shape[0])
        # Sample kernel parameters
        if self.kernel_prior:
            kernel_params = self.kernel_prior()
        else:
            kernel_params = self._sample_kernel_params()
        # Sample noise
        if self.noise_prior:  # this will be removed in the future releases
            noise = self.noise_prior()
        else:
            noise = self._sample_noise()
        # Add mean function (if any)
        if self.mean_fn is not None:
            args = [X]
            if self.mean_fn_prior is not None:
                args += [self.mean_fn_prior()]
            f_loc += self.mean_fn(*args).squeeze()
        # compute kernel
        k = self.kernel(
            X, X,
            kernel_params,
            noise,
            **kwargs
        )
        # sample y according to the standard Gaussian process formula
        numpyro.sample(
            "y",
            dist.MultivariateNormal(loc=f_loc, covariance_matrix=k),
            obs=y,
        )

    def fit(self, rng_key: jnp.array, X: jnp.ndarray, y: jnp.ndarray,
            num_warmup: int = 2000, num_samples: int = 2000,
            num_chains: int = 1, chain_method: str = 'sequential',
            progress_bar: bool = True, print_summary: bool = True,
            device: Type[jaxlib.xla_extension.Device] = None,
            **kwargs: float
            ) -> None:
        """
        Run Hamiltonian Monter Carlo to infer the GP parameters

        Args:
            rng_key: random number generator key
            X: 2D feature vector
            y: 1D target vector
            num_warmup: number of HMC warmup states
            num_samples: number of HMC samples
            num_chains: number of HMC chains
            chain_method: 'sequential', 'parallel' or 'vectorized'
            progress_bar: show progress bar
            print_summary: print summary at the end of sampling
            device:
                optionally specify a cpu or gpu device on which to run the inference;
                e.g., ``device=jax.devices("cpu")[0]`` 
            **jitter:
                Small positive term added to the diagonal part of a covariance
                matrix for numerical stability (Default: 1e-6)
        """
        X, y = self._set_data(X, y)
        if device:
            X = jax.device_put(X, device)
            y = jax.device_put(y, device)
        self.X_train = X
        self.y_train = y

        init_strategy = init_to_median(num_samples=10)
        kernel = NUTS(self.model, init_strategy=init_strategy)
        self.mcmc = MCMC(
            kernel,
            num_warmup=num_warmup,
            num_samples=num_samples,
            num_chains=num_chains,
            chain_method=chain_method,
            progress_bar=progress_bar,
            jit_model_args=False
        )
        self.mcmc.run(rng_key, X, y, **kwargs)
        if print_summary:
            self._print_summary()

    def _sample_noise(self) -> jnp.ndarray:
        if self.noise_prior_dist is not None:
            noise_dist = self.noise_prior_dist
        else:
            noise_dist = dist.LogNormal(0, 1)
        return numpyro.sample("noise", noise_dist)

    def _sample_kernel_params(self, output_scale=True) -> Dict[str, jnp.ndarray]:
        """
        Sample kernel parameters with default
        weakly-informative log-normal priors
        """
        if self.lengthscale_prior_dist is not None:
            length_dist = self.lengthscale_prior_dist
        else:
            length_dist = dist.LogNormal(0.0, 1.0)
        with numpyro.plate('ard', self.kernel_dim):  # allows using ARD kernel for kernel_dim > 1
            length = numpyro.sample("k_length", length_dist)
        if output_scale:
            scale = numpyro.sample("k_scale", dist.LogNormal(0.0, 1.0))
        else:
            scale = numpyro.deterministic("k_scale", jnp.array(1.0))
        if self.kernel_name == 'Periodic':
            period = numpyro.sample("period", dist.LogNormal(0.0, 1.0))
        kernel_params = {
            "k_length": length, "k_scale": scale,
            "period": period if self.kernel_name == "Periodic" else None}
        return kernel_params

    def get_samples(self, chain_dim: bool = False) -> Dict[str, jnp.ndarray]:
        """Get posterior samples (after running the MCMC chains)"""
        return self.mcmc.get_samples(group_by_chain=chain_dim)

    #@partial(jit, static_argnames='self')
    def get_mvn_posterior(self,
                          X_new: jnp.ndarray, params: Dict[str, jnp.ndarray],
                          noiseless: bool = False, **kwargs: float
                          ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Returns parameters (mean and cov) of multivariate normal posterior
        for a single sample of GP parameters
        """
        noise = params["noise"]
        noise_p = noise * (1 - jnp.array(noiseless, int))
        y_residual = self.y_train.copy()
        if self.mean_fn is not None:
            args = [self.X_train, params] if self.mean_fn_prior else [self.X_train]
            y_residual -= self.mean_fn(*args).squeeze()
        # compute kernel matrices for train and test data
        k_pp = self.kernel(X_new, X_new, params, noise_p, **kwargs)
        k_pX = self.kernel(X_new, self.X_train, params, jitter=0.0)
        k_XX = self.kernel(self.X_train, self.X_train, params, noise, **kwargs)
        # compute the predictive covariance and mean
        K_xx_inv = jnp.linalg.inv(k_XX)
        cov = k_pp - jnp.matmul(k_pX, jnp.matmul(K_xx_inv, jnp.transpose(k_pX)))
        mean = jnp.matmul(k_pX, jnp.matmul(K_xx_inv, y_residual))
        if self.mean_fn is not None:
            args = [X_new, params] if self.mean_fn_prior else [X_new]
            mean += self.mean_fn(*args).squeeze()
        return mean, cov

    def _predict(self, rng_key: jnp.ndarray, X_new: jnp.ndarray,
                 params: Dict[str, jnp.ndarray], n: int, noiseless: bool = False,
                 **kwargs: float) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """Prediction with a single sample of GP parameters"""
        # Get the predictive mean and covariance
        y_mean, K = self.get_mvn_posterior(X_new, params, noiseless, **kwargs)
        # draw samples from the posterior predictive for a given set of parameters
        y_sampled = dist.MultivariateNormal(y_mean, K).sample(rng_key, sample_shape=(n,))
        return y_mean, y_sampled

    def _predict_in_batches(self, rng_key: jnp.ndarray,
                            X_new: jnp.ndarray,  batch_size: int = 100,
                            batch_dim: int = 0,
                            samples: Optional[Dict[str, jnp.ndarray]] = None,
                            n: int = 1, filter_nans: bool = False,
                            predict_fn: Callable[[jnp.ndarray, int], Tuple[jnp.ndarray]] = None,
                            noiseless: bool = False,
                            device: Type[jaxlib.xla_extension.Device] = None,
                            **kwargs: float
                            ) -> Tuple[jnp.ndarray, jnp.ndarray]:

        if predict_fn is None:
            predict_fn = lambda xi:  self.predict(
                rng_key, xi, samples, n, filter_nans, noiseless, device, **kwargs)

        def predict_batch(Xi):
            out1, out2 = predict_fn(Xi)
            out1 = jax.device_put(out1, jax.devices("cpu")[0])
            out2 = jax.device_put(out2, jax.devices("cpu")[0])
            return out1, out2

        y_out1, y_out2 = [], []
        for Xi in split_in_batches(X_new, batch_size, dim=batch_dim):
            out1, out2 = predict_batch(Xi)
            y_out1.append(out1)
            y_out2.append(out2)
        return y_out1, y_out2

    def predict_in_batches(self, rng_key: jnp.ndarray,
                           X_new: jnp.ndarray,  batch_size: int = 100,
                           samples: Optional[Dict[str, jnp.ndarray]] = None,
                           n: int = 1, filter_nans: bool = False,
                           predict_fn: Callable[[jnp.ndarray, int], Tuple[jnp.ndarray]] = None,
                           noiseless: bool = False,
                           device: Type[jaxlib.xla_extension.Device] = None,
                           **kwargs: float
                           ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Make prediction at X_new with sampled GP parameters
        by spitting the input array into chunks ("batches") and running
        predict_fn (defaults to self.predict) on each of them one-by-one
        to avoid a memory overflow
        """
        y_pred, y_sampled = self._predict_in_batches(
            rng_key, X_new, batch_size, 0, samples, n,
            filter_nans, predict_fn, noiseless, device, **kwargs)
        y_pred = jnp.concatenate(y_pred, 0)
        y_sampled = jnp.concatenate(y_sampled, -1)
        return y_pred, y_sampled

    def predict(self, rng_key: jnp.ndarray, X_new: jnp.ndarray,
                samples: Optional[Dict[str, jnp.ndarray]] = None,
                n: int = 1, filter_nans: bool = False, noiseless: bool = False,
                device: Type[jaxlib.xla_extension.Device] = None, **kwargs: float
                ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Make prediction at X_new points using posterior samples for GP parameters

        Args:
            rng_key: random number generator key
            X_new: new inputs with *(number of points, number of features)* dimensions
            samples: optional (different) samples with GP parameters
            n: number of samples from Multivariate Normal posterior for each HMC sample with GP parameters
            filter_nans: filter out samples containing NaN values (if any)
            noiseless:
                Noise-free prediction. It is set to False by default as new/unseen data is assumed
                to follow the same distribution as the training data. Hence, since we introduce a model noise
                by default for the training data, we also want to include that noise in our prediction.
            device:
                optionally specify a cpu or gpu device on which to make a prediction;
                e.g., ```device=jax.devices("gpu")[0]```
            **jitter:
                Small positive term added to the diagonal part of a covariance
                matrix for numerical stability (Default: 1e-6)

        Returns
            Center of the mass of sampled means and all the sampled predictions
        """
        X_new = self._set_data(X_new)
        if samples is None:
            samples = self.get_samples(chain_dim=False)
        if device:
            self._set_training_data(device=device)
            X_new = jax.device_put(X_new, device)
            samples = jax.device_put(samples, device)
        num_samples = samples["noise"].shape[0]
        vmap_args = (jra.split(rng_key, num_samples), samples)
        predictive = jax.vmap(
            lambda prms: self._predict(prms[0], X_new, prms[1], n, noiseless, **kwargs))
        y_means, y_sampled = predictive(vmap_args)
        if filter_nans:
            y_sampled_ = [y_i for y_i in y_sampled if not jnp.isnan(y_i).any()]
            y_sampled = jnp.array(y_sampled_)
        return y_means.mean(0), y_sampled

    def sample_from_prior(self, rng_key: jnp.ndarray,
                          X: jnp.ndarray, num_samples: int = 10):
        """
        Samples from prior predictive distribution at X
        """
        X = self._set_data(X)
        prior_predictive = Predictive(self.model, num_samples=num_samples)
        samples = prior_predictive(rng_key, X)
        return samples['y']

    def _set_data(self,
                  X: jnp.ndarray,
                  y: Optional[jnp.ndarray] = None
                  ) -> Union[Tuple[jnp.ndarray], jnp.ndarray]:
        X = X if X.ndim > 1 else X[:, None]
        if y is not None:
            return X, y.squeeze()
        return X

    def _set_training_data(self,
                           X_train_new: jnp.ndarray = None,
                           y_train_new: jnp.ndarray = None,
                           device: Type[jaxlib.xla_extension.Device] = None
                           ) -> None:
        X_train = self.X_train if X_train_new is None else X_train_new
        y_train = self.y_train if y_train_new is None else y_train_new
        if device:
            X_train = jax.device_put(X_train, device)
            y_train = jax.device_put(y_train, device)
        self.X_train = X_train
        self.y_train = y_train

    def _print_summary(self):
        samples = self.get_samples(1)
        numpyro.diagnostics.print_summary(samples)



"""
base_acq.py
==============

Base acquisition functions

Created by Maxim Ziatdinov (email: maxim.ziatdinov@ai4microscopy.com)
"""

from typing import Type, Dict, Optional, Tuple

import jax
import jax.numpy as jnp
import numpyro.distributions as dist

from ..models.gp import ExactGP
from ..utils import get_keys


def ei(moments: Tuple[jnp.ndarray, jnp.ndarray],
       best_f: float = None,
       maximize: bool = False,
       **kwargs) -> jnp.ndarray:
    r"""
    Expected Improvement

    Given a probabilistic model :math:`m` that models the objective function :math:`f`,
    the Expected Improvement at an input point :math:`x` is defined as:

    .. math::
        EI(x) =
        \begin{cases}
        (\mu(x) - f^+ - \xi) \Phi(Z) + \sigma(x) \phi(Z) & \text{if } \sigma(x) > 0 \\
        0 & \text{if } \sigma(x) = 0
        \end{cases}

    where:
    - :math:`\mu(x)` is the predictive mean.
    - :math:`\sigma(x)` is the predictive standard deviation.
    - :math:`f^+` is the value of the best observed sample.
    - :math:`\xi` is a small positive "jitter" term (not used in this function).
    - :math:`Z` is defined as:

    .. math::

        Z = \frac{\mu(x) - f^+ - \xi}{\sigma(x)}

    provided :math:`\sigma(x) > 0`.

    Args:
        moments:
            Tuple with predictive mean and variance
            (first and second moments of predictive distribution).
        best_f:
            Best function value observed so far. Derived from the predictive mean
            when not provided by a user.
        maximize:
            If True, assumes that BO is solving maximization problem.
    """
    mean, var = moments
    if best_f is None:
        best_f = mean.max() if maximize else mean.min()
    sigma = jnp.sqrt(var)
    u = (mean - best_f) / sigma
    if not maximize:
        u = -u
    normal = dist.Normal(jnp.zeros_like(u), jnp.ones_like(u))
    ucdf = normal.cdf(u)
    updf = jnp.exp(normal.log_prob(u))
    acq = sigma * (updf + u * ucdf)
    return acq


def ucb(moments: Tuple[jnp.ndarray, jnp.ndarray],
        beta: float = 0.25,
        maximize: bool = False,
        **kwargs) -> jnp.ndarray:
    r"""
    Upper confidence bound

    Given a probabilistic model :math:`m` that models the objective function :math:`f`,
    the Upper Confidence Bound (UCB) at an input point :math:`x` is defined as:

    .. math::

        UCB(x) = \mu(x) + \kappa \sigma(x)

    where:
    - :math:`\mu(x)` is the predictive mean.
    - :math:`\sigma(x)` is the predictive standard deviation.
    - :math:`\kappa` is the exploration-exploitation trade-off parameter.

    Args:
        moments:
            Tuple with predictive mean and variance
            (first and second moments of predictive distribution).
        maximize: If True, assumes that BO is solving maximization problem
        beta: coefficient balancing exploration-exploitation trade-off
    """
    mean, var = moments
    delta = jnp.sqrt(beta * var)
    if maximize:
        acq = mean + delta
    else:
        acq = -(mean - delta)  # return a negative acq for argmax in BO
    return acq


def ue(moments: Tuple[jnp.ndarray, jnp.ndarray], **kwargs) -> jnp.ndarray:
    r"""
    Uncertainty-based exploration

    Given a probabilistic model :math:`m` that models the objective function :math:`f`,
    the Uncertainty-based Exploration (UE) at an input point :math:`x` targets regions where the model's predictions are most uncertain.
    It quantifies this uncertainty as:

    .. math::

        UE(x) = \sigma^2(x)

    where:
    - :math:`\sigma^2(x)` is the predictive variance of the model at the input point :math:`x`.

    Args:
        moments:
            Tuple with predictive mean and variance
            (first and second moments of predictive distribution).

    """
    _, var = moments
    return jnp.sqrt(var)


def poi(moments: Tuple[jnp.ndarray, jnp.ndarray],
        best_f: float = None, xi: float = 0.01,
        maximize: bool = False, **kwargs) -> jnp.ndarray:
    r"""
    Probability of Improvement

    Args:
        moments:
            Tuple with predictive mean and variance
            (first and second moments of predictive distribution).
        maximize: If True, assumes that BO is solving maximization problem
        xi: Exploration-exploitation trade-off parameter (Defaults to 0.01)
    """
    mean, var = moments
    if best_f is None:
        best_f = mean.max() if maximize else mean.min()
    sigma = jnp.sqrt(var)
    u = (mean - best_f - xi) / sigma
    if not maximize:
        u = -u
    normal = dist.Normal(jnp.zeros_like(u), jnp.ones_like(u))
    return normal.cdf(u)


def kg(model: Type[ExactGP],
       X_new: jnp.ndarray,
       sample: Dict[str, jnp.ndarray],
       rng_key: Optional[jnp.ndarray] = None,
       n: int = 10,
       maximize: bool = True,
       noiseless: bool = True,
       **kwargs):
    r"""
    Knowledge gradient
    
    Given a probabilistic model :math:`m` that models the objective function :math:`f`,
    the Knowledge Gradient (KG) at an input point :math:`x` quantifies the expected improvement in the optimal decision after observing the function value at :math:`x`.

    The KG value is defined as:

    .. math::

        KG(x) = \mathbb{E}[V_{n+1}^* - V_n^* | x]

    where:
    - :math:`V_{n+1}^*` is the optimal expected value of the objective function after \(n+1\) observations.
    - :math:`V_n^*` is the optimal expected value of the objective function based on the current \(n\) observations.

    Args:
        model: trained model
        X: new inputs with shape (N, D), where D is a feature dimension
        sample: a single sample with model parameters
        n: Number fo simulated samples (Defaults to 10)
        maximize: If True, assumes that BO is solving maximization problem
        noiseless:
            Noise-free prediction. It is set to False by default as new/unseen data is assumed
            to follow the same distribution as the training data. Hence, since we introduce a model noise
            for the training data, we also want to include that noise in our prediction.
        rng_key: random number generator key
        **jitter:
            Small positive term added to the diagonal part of a covariance
            matrix for numerical stability (Default: 1e-6)
    """

    if rng_key is None:
        rng_key = get_keys()[0]
    if not isinstance(sample, (tuple, list)):
        sample = (sample,)

    X_train_o = model.X_train.copy()
    y_train_o = model.y_train.copy()

    def kg_for_one_point(x_aug, y_aug, mean_o):
        # Update GP model with augmented data (as if y_sim was an actual observation at x)
        model._set_training_data(x_aug, y_aug)
        # Re-evaluate posterior predictive distribution on all the candidate ("test") points
        mean_aug, _ = model.get_mvn_posterior(X_new, *sample, noiseless=noiseless, **kwargs)
        # Find the maximum mean value
        y_fant = mean_aug.max() if maximize else mean_aug.min()
        # Compute adn return the improvement compared to the original maximum mean value
        mean_o_best = mean_o.max() if maximize else mean_o.min()
        u = y_fant - mean_o_best
        if not maximize:
            u = -u
        return u

    # Get posterior distribution for candidate points
    mean, cov = model.get_mvn_posterior(X_new, *sample, noiseless=noiseless, **kwargs)
    # Simulate potential observations
    y_sim = dist.MultivariateNormal(mean, cov).sample(rng_key, sample_shape=(n,))
    # Augment training data with simulated observations
    X_train_aug = jnp.array([jnp.concatenate([X_train_o, x[None]], axis=0) for x in X_new])
    y_train_aug = []
    for ys in y_sim:
        y_train_aug.append(jnp.array([jnp.concatenate([y_train_o, y[None]]) for y in ys]))
    y_train_aug = jnp.array(y_train_aug)
    # Compute KG
    vectorized_kg = jax.vmap(jax.vmap(kg_for_one_point, in_axes=(0, 0, None)), in_axes=(None, 0, None))
    kg_values = vectorized_kg(X_train_aug, y_train_aug, mean)

    # Reset training data to the original
    model._set_training_data(X_train_o, y_train_o)

    return kg_values.mean(0)



"""
acquisition.py
==============

Acquisition functions

Created by Maxim Ziatdinov (email: maxim.ziatdinov@ai4microscopy.com)
"""

from typing import Type, Optional, Tuple

import jax.numpy as jnp
import jax.random as jra
from jax import vmap
import numpy as onp

from ..models.gp import ExactGP
from .base_acq import ei, ucb, poi, ue, kg
from .penalties import compute_penalty


def _compute_mean_and_var(
        rng_key: jnp.ndarray, model: Type[ExactGP], X: jnp.ndarray,
        n: int, noiseless: bool, **kwargs) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """
    Computes predictive mean and variance
    """
    if model.mcmc is not None:
        _, y_sampled = model.predict(
            rng_key, X, n=n, noiseless=noiseless, **kwargs)
        y_sampled = y_sampled.reshape(n * y_sampled.shape[0], -1)
        mean, var = y_sampled.mean(0), y_sampled.var(0)
    else:
        mean, var = model.predict(rng_key, X, noiseless=noiseless, **kwargs)
    return mean, var


def _compute_penalties(
        X: jnp.ndarray, recent_points: jnp.ndarray, penalty: str,
        penalty_factor: float, grid_indices: jnp.ndarray) -> jnp.ndarray:
    """
    Computes penaltes for recent points to be substracted
    from acqusition function values
    """
    X_ = grid_indices if grid_indices is not None else X
    return compute_penalty(X_, recent_points, penalty, penalty_factor)


def EI(rng_key: jnp.ndarray, model: Type[ExactGP],
       X: jnp.ndarray, best_f: float = None,
       maximize: bool = False, n: int = 1,
       noiseless: bool = False,
       penalty: Optional[str] = None,
       recent_points: jnp.ndarray = None,
       grid_indices: jnp.ndarray = None,
       penalty_factor: float = 1.0,
       **kwargs) -> jnp.ndarray:
    r"""
    Expected Improvement

    Given a probabilistic model :math:`m` that models the objective function :math:`f`,
    the Expected Improvement at an input point :math:`x` is defined as:

    .. math::
        EI(x) =
        \begin{cases}
        (\mu(x) - f^+) \Phi(Z) + \sigma(x) \phi(Z) & \text{if } \sigma(x) > 0 \\
        0 & \text{if } \sigma(x) = 0
        \end{cases}

    where :math:`\mu(x)` is the predictive mean, :math:`\sigma(x)` is the predictive standard deviation,
    :math:`f^+` is the value of the best observed sample. :math:`Z` is defined as:

    .. math::

        Z = \frac{\mu(x) - f^+}{\sigma(x)}

    provided :math:`\sigma(x) > 0`.

    In the case of HMC, the function leverages multiple predictive posteriors, each associated
    with a different HMC sample of the GP model parameters, to capture both prediction uncertainty
    and hyperparameter uncertainty. In this setup, the uncertainty in parameters of probabilistic
    mean function (if any) also contributes to the acquisition function values.

    Args:
        rng_key: JAX random number generator key
        model: trained model
        X: new inputs
        best_f:
            Best function value observed so far. Derived from the predictive mean
            when not provided by a user.
        maximize: If True, assumes that BO is solving maximization problem
        n: number of samples drawn from each MVN distribution
           (number of distributions is equal to the number of HMC samples)
        noiseless:
            Noise-free prediction. It is set to False by default as new/unseen data is assumed
            to follow the same distribution as the training data. Hence, since we introduce a model noise
            for the training data, we also want to include that noise in our prediction.
        penalty:
            Penalty applied to the acquisition function to discourage re-evaluation
            at or near points that were recently evaluated. Options are:

            - 'delta':
            The infinite penalty is applied to the recently visited points.

            - 'inverse_distance':
            Modifies the acquisition function by penalizing points near the recent points.

            For the 'inverse_distance', the acqusition function is penalized as:

            .. math::
                \alpha - \lambda \cdot \pi(X, r)

            where :math:`\pi(X, r)` computes a penalty for points in :math:`X` based on their distance to recent points :math:`r`,
            :math:`\alpha` represents the acquisition function, and :math:`\lambda` represents the penalty factor.
        recent_points:
            An array of recently visited points [oldest, ..., newest] provided by user
        grid_indices:
            Grid indices of data points in X array for the penalty term calculation.
            For example, if each data point is an image patch, the indices could correspond
            to the (i, j) pixel coordinates of their centers in the original image.
        penalty_factor:
            Penalty factor :math:`\lambda` in :math:`\alpha - \lambda \cdot \pi(X, r)`
        **jitter:
            Small positive term added to the diagonal part of a covariance
            matrix for numerical stability (Default: 1e-6)
    """
    if penalty and not isinstance(recent_points, (onp.ndarray, jnp.ndarray)):
        raise ValueError("Please provide an array of recently visited points")

    X = X[:, None] if X.ndim < 2 else X

    moments = _compute_mean_and_var(rng_key, model, X, n, noiseless, **kwargs)

    acq = ei(moments, best_f, maximize)

    if penalty:
        acq -= _compute_penalties(X, recent_points, penalty, penalty_factor, grid_indices)

    return acq


def UCB(rng_key: jnp.ndarray, model: Type[ExactGP],
        X: jnp.ndarray, beta: float = .25,
        maximize: bool = False, n: int = 1,
        noiseless: bool = False,
        penalty: Optional[str] = None,
        recent_points: jnp.ndarray = None,
        grid_indices: jnp.ndarray = None,
        penalty_factor: float = 1.0,
        **kwargs) -> jnp.ndarray:
    r"""
    Upper confidence bound

    Given a probabilistic model :math:`m` that models the objective function :math:`f`,
    the Upper Confidence Bound at an input point :math:`x` is defined as:

    .. math::

        UCB(x) = \mu(x) + \kappa \sigma(x)

    where :math:`\mu(x)` is the predictive mean, :math:`\sigma(x)` is the predictive standard deviation,
    and :math:`\kappa` is the exploration-exploitation trade-off parameter.

    In the case of HMC, the function leverages multiple predictive posteriors, each associated
    with a different HMC sample of the GP model parameters, to capture both prediction uncertainty
    and hyperparameter uncertainty. In this setup, the uncertainty in parameters of probabilistic
    mean function (if any) also contributes to the acquisition function values.

    Args:
        rng_key: JAX random number generator key
        model: trained model
        X: new inputs
        beta: coefficient balancing exploration-exploitation trade-off
        maximize: If True, assumes that BO is solving maximization problem
        n: number of samples drawn from each MVN distribution
           (number of distributions is equal to the number of HMC samples)
        noiseless:
            Noise-free prediction. It is set to False by default as new/unseen data is assumed
            to follow the same distribution as the training data. Hence, since we introduce a model noise
            for the training data, we also want to include that noise in our prediction.
        penalty:
            Penalty applied to the acquisition function to discourage re-evaluation
            at or near points that were recently evaluated. Options are:

            - 'delta':
            The infinite penalty is applied to the recently visited points.

            - 'inverse_distance':
            Modifies the acquisition function by penalizing points near the recent points.

            For the 'inverse_distance', the acqusition function is penalized as:

            .. math::
                \alpha - \lambda \cdot \pi(X, r)

            where :math:`\pi(X, r)` computes a penalty for points in :math:`X` based on their distance to recent points :math:`r`,
            :math:`\alpha` represents the acquisition function, and :math:`\lambda` represents the penalty factor.
        recent_points:
            An array of recently visited points [oldest, ..., newest] provided by user
        grid_indices:
            Grid indices of data points in X array for the penalty term calculation.
            For example, if each data point is an image patch, the indices could correspond
            to the (i, j) pixel coordinates of their centers in the original image.
        penalty_factor:
            Penalty factor :math:`\lambda` in :math:`\alpha - \lambda \cdot \pi(X, r)`
        **jitter:
            Small positive term added to the diagonal part of a covariance
            matrix for numerical stability (Default: 1e-6)
    """

    if penalty and not isinstance(recent_points, (onp.ndarray, jnp.ndarray)):
        raise ValueError("Please provide an array of recently visited points")

    X = X[:, None] if X.ndim < 2 else X

    moments = _compute_mean_and_var(rng_key, model, X, n, noiseless, **kwargs)

    acq = ucb(moments, beta, maximize)

    if penalty:
        acq -= _compute_penalties(X, recent_points, penalty, penalty_factor, grid_indices)

    return acq


def POI(rng_key: jnp.ndarray, model: Type[ExactGP],
        X: jnp.ndarray, best_f: float = None,
        xi: float = 0.01, maximize: bool = False,
        n: int = 1, noiseless: bool = False,
        penalty: Optional[str] = None,
        recent_points: jnp.ndarray = None,
        grid_indices: jnp.ndarray = None,
        penalty_factor: float = 1.0,
        **kwargs) -> jnp.ndarray:
    r"""
    Probability of Improvement

    Given a probabilistic model :math:`m` that models the objective function :math:`f`,
    the Probability of Improvement at an input point :math:`x` is defined as:

    .. math::

        PI(x) = \Phi\left(\frac{\mu(x) - f^+ - \xi}{\sigma(x)}\right)

    where :math:`\mu(x)` is the predictive mean, :math:`\sigma(x)` is the predictive standard deviation,
    :math:`f^+` is the value of the best observed sample, :math:`\xi` is a small positive "jitter" term to encourage more exploration,
    and :math:`\Phi` is the cumulative distribution function (CDF) of the standard normal distribution.

    In the case of HMC, the function leverages multiple predictive posteriors, each associated
    with a different HMC sample of the GP model parameters, to capture both prediction uncertainty
    and hyperparameter uncertainty. In this setup, the uncertainty in parameters of probabilistic
    mean function (if any) also contributes to the acquisition function values.

    Args:
        rng_key: JAX random number generator key
        model: trained model
        X: new inputs
        best_f:
            Best function value observed so far. Derived from the predictive mean
            when not provided by a user.
        xi: coefficient affecting exploration-exploitation trade-off
        maximize: If True, assumes that BO is solving maximization problem
        n: number of samples drawn from each MVN distribution
           (number of distributions is equal to the number of HMC samples)
        noiseless:
            Noise-free prediction. It is set to False by default as new/unseen data is assumed
            to follow the same distribution as the training data. Hence, since we introduce a model noise
            for the training data, we also want to include that noise in our prediction.
        penalty:
            Penalty applied to the acquisition function to discourage re-evaluation
            at or near points that were recently evaluated. Options are:

            - 'delta':
            The infinite penalty is applied to the recently visited points.

            - 'inverse_distance':
            Modifies the acquisition function by penalizing points near the recent points.

            For the 'inverse_distance', the acqusition function is penalized as:

            .. math::
                \alpha - \lambda \cdot \pi(X, r)

            where :math:`\pi(X, r)` computes a penalty for points in :math:`X` based on their distance to recent points :math:`r`,
            :math:`\alpha` represents the acquisition function, and :math:`\lambda` represents the penalty factor.
        recent_points:
            An array of recently visited points [oldest, ..., newest] provided by user
        grid_indices:
            Grid indices of data points in X array for the penalty term calculation.
            For example, if each data point is an image patch, the indices could correspond
            to the (i, j) pixel coordinates of their centers in the original image.
        penalty_factor:
            Penalty factor :math:`\lambda` in :math:`\alpha - \lambda \cdot \pi(X, r)`
        **jitter:
            Small positive term added to the diagonal part of a covariance
            matrix for numerical stability (Default: 1e-6)
    """
    if penalty and not isinstance(recent_points, (onp.ndarray, jnp.ndarray)):
        raise ValueError("Please provide an array of recently visited points")

    X = X[:, None] if X.ndim < 2 else X

    moments = _compute_mean_and_var(rng_key, model, X, n, noiseless, **kwargs)

    acq = poi(moments, best_f, xi, maximize)

    if penalty:
        acq -= _compute_penalties(X, recent_points, penalty, penalty_factor, grid_indices)

    return acq


def UE(rng_key: jnp.ndarray, model: Type[ExactGP],
        X: jnp.ndarray,
        n: int = 1,
        noiseless: bool = False,
        penalty: Optional[str] = None,
        recent_points: jnp.ndarray = None,
        grid_indices: jnp.ndarray = None,
        penalty_factor: float = 1.0,
        **kwargs) -> jnp.ndarray:

    r"""
    Uncertainty-based exploration

    Given a probabilistic model :math:`m` that models the objective function :math:`f`,
    the Uncertainty-based Exploration (UE) at an input point :math:`x` targets regions where the model's predictions are most uncertain.
    It quantifies this uncertainty as:

    .. math::

        UE(x) = \sigma^2(x)

    where :math:`\sigma^2(x)` is the predictive variance of the model at the input point :math:`x`.

    In the case of HMC, the function leverages multiple predictive posteriors, each associated
    with a different HMC sample of the GP model parameters, to capture both prediction uncertainty
    and hyperparameter uncertainty. In this setup, the uncertainty in parameters of probabilistic
    mean function (if any) also contributes to the acquisition function values.

    Args:
        rng_key: JAX random number generator key
        model: trained model
        X: new inputs
        n: number of samples drawn from each MVN distribution
           (number of distributions is equal to the number of HMC samples)
        noiseless:
            Noise-free prediction. It is set to False by default as new/unseen data is assumed
            to follow the same distribution as the training data. Hence, since we introduce a model noise
            for the training data, we also want to include that noise in our prediction.
        penalty:
            Penalty applied to the acquisition function to discourage re-evaluation
            at or near points that were recently evaluated. Options are:

            - 'delta':
            The infinite penalty is applied to the recently visited points.

            - 'inverse_distance':
            Modifies the acquisition function by penalizing points near the recent points.

            For the 'inverse_distance', the acqusition function is penalized as:

            .. math::
                \alpha - \lambda \cdot \pi(X, r)

            where :math:`\pi(X, r)` computes a penalty for points in :math:`X` based on their distance to recent points :math:`r`,
            :math:`\alpha` represents the acquisition function, and :math:`\lambda` represents the penalty factor.
        recent_points:
            An array of recently visited points [oldest, ..., newest] provided by user
        grid_indices:
            Grid indices of data points in X array for the penalty term calculation.
            For example, if each data point is an image patch, the indices could correspond
            to the (i, j) pixel coordinates of their centers in the original image.
        penalty_factor:
            Penalty factor :math:`\lambda` in :math:`\alpha - \lambda \cdot \pi(X, r)`
        **jitter:
            Small positive term added to the diagonal part of a covariance
            matrix for numerical stability (Default: 1e-6)
    """
    if penalty and not isinstance(recent_points, (onp.ndarray, jnp.ndarray)):
        raise ValueError("Please provide an array of recently visited points")
    X = X[:, None] if X.ndim < 2 else X

    moments = _compute_mean_and_var(rng_key, model, X, n, noiseless, **kwargs)

    acq = ue(moments)

    if penalty:
        X_ = grid_indices if grid_indices is not None else X
        penalties = compute_penalty(X_, recent_points, penalty, penalty_factor)

        acq -= penalties
    return acq


def KG(rng_key: jnp.ndarray,
       model: Type[ExactGP],
       X: jnp.ndarray,
       n: int = 1,
       maximize: bool = False,
       noiseless: bool = False,
       penalty: Optional[str] = None,
       recent_points: jnp.ndarray = None,
       grid_indices: jnp.ndarray = None,
       penalty_factor: float = 1.0,
       **kwargs) -> jnp.ndarray:
    r"""
    Knowledge gradient

    Given a probabilistic model :math:`m` that models the objective function :math:`f`,
    the Knowledge Gradient (KG) at an input point :math:`x` quantifies the expected improvement
    in the optimal decision after observing the function value at :math:`x`.

    The KG value is defined as:

    .. math::

        KG(x) = \mathbb{E}[V_{n+1}^* - V_n^* | x]

    where :math:`V_{n+1}^*` is the optimal expected value of the objective function after \(n+1\) observations and
    :math:`V_n^*` is the optimal expected value of the objective function based on the current \(n\) observations.

    Args:
        rng_key:
            JAX random number generator key for sampling simulated observations
        model:
            Trained model
        X:
            New inputs
        n:
            Number of simulated samples for each point in X
        maximize:
            If True, assumes that BO is solving maximization problem
        noiseless:
            Noise-free prediction. It is set to False by default as new/unseen data is assumed
            to follow the same distribution as the training data. Hence, since we introduce a model noise
            for the training data, we also want to include that noise in our prediction.
        penalty:
            Penalty applied to the acquisition function to discourage re-evaluation
            at or near points that were recently evaluated. Options are:

            - 'delta':
            The infinite penalty is applied to the recently visited points.

            - 'inverse_distance':
            Modifies the acquisition function by penalizing points near the recent points.

            For the 'inverse_distance', the acqusition function is penalized as:

            .. math::
                \alpha - \lambda \cdot \pi(X, r)

            where :math:`\pi(X, r)` computes a penalty for points in :math:`X` based on their distance to recent points :math:`r`,
            :math:`\alpha` represents the acquisition function, and :math:`\lambda` represents the penalty factor.
        recent_points:
            An array of recently visited points [oldest, ..., newest] provided by user
        grid_indices:
            Grid indices of data points in X array for the penalty term calculation.
            For example, if each data point is an image patch, the indices could correspond
            to the (i, j) pixel coordinates of their centers in the original image.
        penalty_factor:
            Penalty factor :math:`\lambda` in :math:`\alpha - \lambda \cdot \pi(X, r)`
        **jitter:
            Small positive term added to the diagonal part of a covariance
            matrix for numerical stability (Default: 1e-6)
    """
    if penalty and not isinstance(recent_points, (onp.ndarray, jnp.ndarray)):
        raise ValueError("Please provide an array of recently visited points")

    X = X[:, None] if X.ndim < 2 else X
    samples = model.get_samples()

    if model.mcmc is None:
        acq = kg(model, X, samples, rng_key, n, maximize, noiseless, **kwargs)
    else:
        vec_kg = vmap(kg, in_axes=(None, None, 0, 0, None, None, None))
        samples = model.get_samples()
        keys = jra.split(rng_key, num=len(next(iter(samples.values()))))
        acq = vec_kg(model, X, samples, keys, n, maximize, noiseless, **kwargs)

    if penalty:
        acq -= _compute_penalties(X, recent_points, penalty, penalty_factor, grid_indices)

    return acq


def Thompson(rng_key: jnp.ndarray,
             model: Type[ExactGP],
             X: jnp.ndarray, n: int = 1,
             noiseless: bool = False,
             **kwargs) -> jnp.ndarray:
    """
    Thompson sampling.

    For MAP approximation, it draws a single sample of a function from the
    posterior predictive distribution. In the case of HMC, it draws a single posterior
    sample from the HMC samples of GP model parameters and then samples a function from it.

    Args:
        rng_key: JAX random number generator key
        model: trained model
        X: new inputs
        n: number of samples drawn from the randomly selected MVN distribution
        noiseless:
            Noise-free prediction. It is set to False by default as new/unseen data is assumed
            to follow the same distribution as the training data. Hence, since we introduce a model noise
            for the training data, we also want to include that noise in our prediction.
        **jitter:
            Small positive term added to the diagonal part of a covariance
            matrix for numerical stability (Default: 1e-6)
    """
    if model.mcmc is not None:
        posterior_samples = model.get_samples()
        idx = jra.randint(rng_key, (1,), 0, len(posterior_samples["k_length"]))
        samples = {k: v[idx] for (k, v) in posterior_samples.items()}
        _, tsample = model.predict(
            rng_key, X, samples, n, noiseless=noiseless, **kwargs)
        if n > 1:
            tsample = tsample.mean(1).squeeze()
    else:
        _, tsample = model.sample_from_posterior(
            rng_key, X, n=1, noiseless=noiseless, **kwargs)
    return tsample



"""
batch_acquisition.py
==============

Batch-mode acquisition functions

Created by Maxim Ziatdinov (email: maxim.ziatdinov@ai4microscopy.com)
"""

from typing import Type, Optional, Callable

import jax.numpy as jnp
from jax import vmap
import jax.random as jra

from ..models.gp import ExactGP
from ..utils import random_sample_dict
from .acquisition import ei, ucb, poi, kg


def _compute_batch_acquisition(
        rng_key: jnp.ndarray,
        model: Type[ExactGP],
        X: jnp.ndarray,
        single_acq_fn: Callable,
        maximize_distance: bool = False,
        subsample_size: int = 1,
        n_evals: int = 10,
        indices: Optional[jnp.ndarray] = None,
        **kwargs) -> jnp.ndarray:
    """Function for computing batch acquisition of a given type"""

    if model.mcmc is None:
        raise ValueError("The model needs to be fully Bayesian")

    X = X[:, None] if X.ndim < 2 else X

    f = vmap(single_acq_fn, in_axes=(0, None))

    if not maximize_distance:
        samples = random_sample_dict(model.get_samples(), subsample_size, rng_key)
        acq = f(samples, X)

    else:
        X_ = jnp.array(indices) if indices is not None else jnp.array(X)

        def compute_acq_and_distance(subkey):
            samples = random_sample_dict(model.get_samples(), subsample_size, subkey)
            acq = f(samples, X_)
            points = acq.argmax(-1)
            d = jnp.linalg.norm(points).mean()
            return acq, d

        subkeys = jra.split(rng_key, num=n_evals)
        acq_all, dist_all = vmap(compute_acq_and_distance)(subkeys)
        idx = dist_all.argmax()
        acq = acq_all[idx]

    return acq


def qEI(rng_key: jnp.ndarray,
        model: Type[ExactGP],
        X: jnp.ndarray,
        best_f: float = None,
        maximize: bool = False,
        noiseless: bool = False,
        maximize_distance: bool = False,
        subsample_size: int = 1,
        n_evals: int = 10,
        indices: Optional[jnp.ndarray] = None,
        **kwargs) -> jnp.ndarray:
    """
    Batch-mode Expected Improvement

    qEI computes the Expected Improvement values for given input points `X` using multiple randomly drawn samples
    from the HMC-inferred model's posterior. If `maximize_distance` is enabled, qEI considers diversity among the
    posterior samples by maximizing the mean distance between samples that give the highest acquisition
    values across multiple evaluations.

    Args:
        rng_key: random number generator key
        model: trained model
        X: new inputs
        best_f:
            Best function value observed so far. Derived from the predictive mean
            when not provided by a user.
        maximize:
            If True, assumes that BO is solving maximization problem
        noiseless:
            Noise-free prediction. It is set to False by default as new/unseen data is assumed
            to follow the same distribution as the training data. Hence, since we introduce a model noise
            for the training data, we also want to include that noise in our prediction.
        maximize_distance:
            If set to True, it means we want our batch to contain points that
            are as far apart as possible in the acquisition function space.
            This encourages diversity in the batch.
        subsample_size:
            Size of the subsample from the GP model's MCMC samples.
        n_evals:
            Number of evaluations (how many times a ramdom subsample is drawn)
            when maximizing distance between maxima of different EIs in a batch.
        indices:
            Indices of the input points.

    Returns:
        The computed batch Expected Improvement values at the provided input points X.
    """

    def single_acq(sample, X):
        mean, cov = model.get_mvn_posterior(X, sample, noiseless, **kwargs)
        return ei((mean, cov.diagonal()), best_f, maximize)

    return _compute_batch_acquisition(
        rng_key, model, X, single_acq, maximize_distance,
        subsample_size, n_evals, indices, **kwargs)


def qUCB(rng_key: jnp.ndarray,
         model: Type[ExactGP],
         X: jnp.ndarray,
         beta: float = 0.25,
         maximize: bool = False,
         noiseless: bool = False,
         maximize_distance: bool = False,
         subsample_size: int = 1,
         n_evals: int = 10,
         indices: Optional[jnp.ndarray] = None,
         **kwargs) -> jnp.ndarray:
    """
    Batch-mode Upper Confidence Bound

    qUCB computes the Upper Confidence Bound values for given input points `X` using multiple randomly drawn samples
    from the HMC-inferred model's posterior. If `maximize_distance` is enabled, qUCB considers diversity among the
    posterior samples by maximizing the mean distance between samples that give the highest acquisition
    values across multiple evaluations.

    Args:
        rng_key: random number generator key
        model: trained model
        X: new inputs
        best_f:
            Best function value observed so far. Derived from the predictive mean
            when not provided by a user.
        maximize:
            If True, assumes that BO is solving maximization problem
        noiseless:
            Noise-free prediction. It is set to False by default as new/unseen data is assumed
            to follow the same distribution as the training data. Hence, since we introduce a model noise
            for the training data, we also want to include that noise in our prediction.
        maximize_distance:
            If set to True, it means we want our batch to contain points that
            are as far apart as possible in the acquisition function space.
            This encourages diversity in the batch.
        subsample_size:
            Size of the subsample from the GP model's MCMC samples.
        n_evals:
            Number of evaluations (how many times a ramdom subsample is drawn)
            when maximizing distance between maxima of different EIs in a batch.
        indices:
            Indices of the input points.

    Returns:
        The computed batch Expected Improvement values at the provided input points X.
    """

    def single_acq(sample, X):
        mean, cov = model.get_mvn_posterior(X, sample, noiseless, **kwargs)
        return ucb((mean, cov.diagonal()), beta, maximize)

    return _compute_batch_acquisition(
        rng_key, model, X, single_acq, maximize_distance,
        subsample_size, n_evals, indices, **kwargs)


def qPOI(rng_key: jnp.ndarray,
         model: Type[ExactGP],
         X: jnp.ndarray,
         best_f: float = None,
         maximize: bool = False,
         noiseless: bool = False,
         maximize_distance: bool = False,
         subsample_size: int = 1,
         n_evals: int = 10,
         indices: Optional[jnp.ndarray] = None,
         **kwargs) -> jnp.ndarray:
    """
    Batch-mode Probability of Improvement

    qPOI computes the Probability of Improvement values for given input points `X` using multiple randomly drawn samples
    from the HMC-inferred model's posterior. If `maximize_distance` is enabled, qPOI considers diversity among the
    posterior samples by maximizing the mean distance between samples that give the highest acquisition
    values across multiple evaluations.

    Args:
        rng_key: random number generator key
        model: trained model
        X: new inputs
        best_f:
            Best function value observed so far. Derived from the predictive mean
            when not provided by a user.
        maximize:
            If True, assumes that BO is solving maximization problem
        noiseless:
            Noise-free prediction. It is set to False by default as new/unseen data is assumed
            to follow the same distribution as the training data. Hence, since we introduce a model noise
            for the training data, we also want to include that noise in our prediction.
        maximize_distance:
            If set to True, it means we want our batch to contain points that
            are as far apart as possible in the acquisition function space.
            This encourages diversity in the batch.
        subsample_size:
            Size of the subsample from the GP model's MCMC samples.
        n_evals:
            Number of evaluations (how many times a ramdom subsample is drawn)
            when maximizing distance between maxima of different EIs in a batch.
        indices:
            Indices of the input points.

    Returns:
        The computed batch Expected Improvement values at the provided input points X.
    """

    def single_acq(sample, X):
        mean, cov = model.get_mvn_posterior(X, sample, noiseless, **kwargs)
        return poi((mean, cov.diagonal()), best_f, maximize)

    return _compute_batch_acquisition(
        rng_key, model, X, single_acq, maximize_distance,
        subsample_size, n_evals, indices, **kwargs)


def qKG(rng_key: jnp.ndarray,
        model: Type[ExactGP],
        X: jnp.ndarray,
        n: int = 10,
        maximize: bool = False,
        noiseless: bool = False,
        maximize_distance: bool = False,
        subsample_size: int = 1,
        n_evals: int = 10,
        indices: Optional[jnp.ndarray] = None,
        **kwargs) -> jnp.ndarray:
    """
    Batch-mode Knowledge Gradient

    qKG computes the Knowledge Gradient values for given input points `X` using multiple randomly drawn samples
    from the HMC-inferred model's posterior. If `maximize_distance` is enabled, qKG considers diversity among the
    posterior samples by maximizing the mean distance between samples that give the highest acquisition
    values across multiple evaluations.

    Args:
        rng_key: random number generator key
        model: trained model
        X: new inputs
        n: number of simulated samples for each point in X
        maximize: If True, assumes that BO is solving maximization problem
        noiseless:
            Noise-free prediction. It is set to False by default as new/unseen data is assumed
            to follow the same distribution as the training data. Hence, since we introduce a model noise
            for the training data, we also want to include that noise in our prediction.
        maximize_distance:
            If set to True, it means we want our batch to contain points that
            are as far apart as possible in the acquisition function space.
            This encourages diversity in the batch.
        subsample_size:
            Size of the subsample from the GP model's MCMC samples.
        n_evals:
            Number of evaluations (how many times a ramdom subsample is drawn)
            when maximizing distance between maxima of different EIs in a batch.
        indices:
            Indices of the input points.

    Returns:
        The computed batch Knowledge Gradient values at the provided input points X.
    """
    def single_acq(sample, X):
        return kg(model, X, sample, rng_key, n, maximize, noiseless, **kwargs)

    return _compute_batch_acquisition(
        rng_key, model, X, single_acq, maximize_distance,
        subsample_size, n_evals, indices, **kwargs)



import sys
import pytest
import numpy as onp
import jax.numpy as jnp
from numpy.testing import assert_equal, assert_

sys.path.insert(0, "../gpax/")

from gpax.kernels import (RBFKernel, MaternKernel, PeriodicKernel,
                          index_kernel, nngp_erf, nngp_relu, NNGPKernel,
                          MultitaskKernel, MultivariateKernel, LCMKernel)


@pytest.mark.parametrize("kernel", [RBFKernel, MaternKernel])
@pytest.mark.parametrize("dim", [1, 2])
def test_data_kernel_shapes(kernel, dim):
    x1 = onp.random.randn(5, dim)
    x2 = onp.random.randn(5, dim)
    params = {"k_length": jnp.array(1.0), "k_scale": jnp.array(1.0)}
    k = kernel(x1, x2, params)
    assert_equal(k.shape, (5, 5))


@pytest.mark.parametrize("dim", [1, 2])
def test_periodkernel_shapes(dim):
    x1 = onp.random.randn(5, dim)
    x2 = onp.random.randn(5, dim)
    params = {"k_length": jnp.array(1.0), "k_scale": jnp.array(1.0), "period": jnp.array(1.0)}
    k = PeriodicKernel(x1, x2, params)
    assert_equal(k.shape, (5, 5))


@pytest.mark.parametrize("kernel", [RBFKernel, MaternKernel])
@pytest.mark.parametrize("dim", [1, 2])
def test_data_kernel_ard_shapes(kernel, dim):
    x1 = onp.random.randn(5, dim)
    x2 = onp.random.randn(5, dim)
    params = {"k_length": jnp.ones(dim), "k_scale": jnp.array(1.0)}
    k = kernel(x1, x2, params)
    assert_equal(k.shape, (5, 5))


def test_index_kernel_shapes():
    indices1 = jnp.array([0, 1, 2])
    indices2 = jnp.array([2, 1, 0])
    params = {"W": jnp.array([[1, 2], [3, 4]]), "v": jnp.array([1, 2])}
    result = index_kernel(indices1, indices2, params)
    assert_(result.shape == (len(indices1), len(indices2)),  "Incorrect shape of result")


def test_index_kernel_shapes_uneven_obs():
    indices1 = jnp.array([1, 2])
    indices2 = jnp.array([2, 1, 0])
    params = {"W": jnp.array([[1, 2], [3, 4]]), "v": jnp.array([1, 2])}
    result = index_kernel(indices1, indices2, params)
    assert_(result.shape == (len(indices1), len(indices2)),  "Incorrect shape of result")


def test_index_kernel_computations():
    indices1 = jnp.array([0, 1])
    indices2 = jnp.array([1, 0])
    params = {"W": jnp.array([[1, 0], [0, 1]]), "v": jnp.array([1, 0])}
    result = index_kernel(indices1, indices2, params)
    expected_result = jnp.array([[0, 2], [1, 0]])
    assert_(jnp.allclose(result, expected_result), "Incorrect computations")


@pytest.mark.parametrize("depth", [1, 2, 3])
@pytest.mark.parametrize("kernel", [nngp_erf, nngp_relu])
@pytest.mark.parametrize("dim", [1, 2])
def test_nngp_shapes(kernel, dim, depth):
    x1 = onp.random.randn(1, dim)
    x2 = onp.random.randn(1, dim)
    var_b = jnp.array(1.0)
    var_w = jnp.array(1.0)
    k = kernel(x1, x2, var_b, var_w, depth)
    assert_equal(k.shape, (1,))


@pytest.mark.parametrize("depth", [1, 2, 3])
@pytest.mark.parametrize("activation", ["erf", "relu"])
@pytest.mark.parametrize("dim", [1, 2])
def test_NNGPKernel(activation, dim, depth):
    x1 = onp.random.randn(5, dim)
    x2 = onp.random.randn(5, dim)
    params = {"var_b": jnp.array(1.0), "var_w": jnp.array(1.0)}
    kernel = NNGPKernel(activation, depth)
    k = kernel(x1, x2, params)
    assert_equal(k.shape, (5, 5))


def test_NNGPKernel_activations():
    x1 = onp.random.randn(5, 1)
    x2 = onp.random.randn(5, 1)
    params = {"var_b": jnp.array(1.0), "var_w": jnp.array(1.0)}
    kernel1 = NNGPKernel(activation='erf')
    k1 = kernel1(x1, x2, params)
    kernel2 = NNGPKernel(activation='relu')
    k2 = kernel2(x1, x2, params)
    assert_(not jnp.allclose(k1, k2, rtol=1e-3))


def test_MultiTaskKernel():
    base_kernel = 'RBF'
    mtkernel = MultitaskKernel(base_kernel)
    assert_(callable(mtkernel), "The result of MultitaskKernel should be a function.")


@pytest.mark.parametrize("data_kernel", [RBFKernel, MaternKernel])
@pytest.mark.parametrize("dim", [1, 2])
def test_multitask_kernel_shapes_test_noiseless(data_kernel, dim):
    x1 = onp.random.randn(5, dim)
    x2 = onp.random.randn(3, dim)
    x1 = onp.column_stack([x1, onp.zeros_like(x1)])
    x2 = onp.column_stack([x2, onp.ones_like(x2)])
    x12 = onp.vstack([x1, x2])
    params = {"k_length": jnp.array(1.0), "k_scale": jnp.array(1.0),
              "W": jnp.array([[1, 0], [0, 1]]), "v": jnp.array([1, 0])}
    mtkernel = MultitaskKernel(data_kernel)
    k = mtkernel(x12, x12, params)
    assert_equal(k.shape, (len(x12), len(x12)))


@pytest.mark.parametrize("data_kernel", [RBFKernel, MaternKernel])
@pytest.mark.parametrize("dim", [1, 2])
def test_multitask_kernel_shapes_test_noisy(data_kernel, dim):
    x1 = onp.random.randn(5, dim)
    x2 = onp.random.randn(3, dim)
    x1 = onp.column_stack([x1, onp.zeros_like(x1)])
    x2 = onp.column_stack([x2, onp.ones_like(x2)])
    x12 = onp.vstack([x1, x2])
    params = {"k_length": jnp.array(1.0), "k_scale": jnp.array(1.0),
              "W": jnp.array([[1, 0], [0, 1]]), "v": jnp.array([1, 0])}
    noise = jnp.array([1.0, 1.0])
    mtkernel = MultitaskKernel(data_kernel)
    k = mtkernel(x12, x12, params, noise)
    assert_equal(k.shape, (len(x12), len(x12)))


@pytest.mark.parametrize("data_kernel", [RBFKernel, MaternKernel])
@pytest.mark.parametrize("dim", [1, 2])
def test_multitask_kernel_shapes_train(data_kernel, dim):
    x1 = onp.random.randn(5, dim)
    x2 = onp.random.randn(5, dim)
    x1 = onp.column_stack([x1, onp.zeros_like(x1)])
    x2 = onp.column_stack([x2, onp.ones_like(x2)])
    x12 = onp.vstack([x1, x2])
    params = {"k_length": jnp.array(1.0), "k_scale": jnp.array(1.0),
              "W": jnp.array([[1, 0], [0, 1]]), "v": jnp.array([1, 0])}
    noise = jnp.array([1.0, 1.0])
    mtkernel = MultitaskKernel(data_kernel)
    k = mtkernel(x12, x12, params, noise)
    assert_equal(k.shape, (len(x12), len(x12)))


def test_MultiVariateKernel():
    base_kernel = 'RBF'
    num_tasks = 2
    mtkernel = MultivariateKernel(base_kernel, num_tasks)
    assert_(callable(mtkernel), "The result of MultiVariateKernel should be a function.")


@pytest.mark.parametrize("data_kernel", [RBFKernel, MaternKernel])
@pytest.mark.parametrize("num_tasks", [2, 3])
@pytest.mark.parametrize("rank", [1, 2])
@pytest.mark.parametrize("dim", [1, 2])
def test_multivariate_kernel_shapes_test_noisy(data_kernel, dim, num_tasks, rank):
    x1 = onp.random.randn(5, dim)
    x2 = onp.random.randn(3, dim)
    params = {"k_length": jnp.array(1.0), "k_scale": jnp.array(1.0),
              "W": jnp.ones((num_tasks, rank)), "v": jnp.ones(num_tasks)}
    noise = jnp.ones(num_tasks)
    mtkernel = MultivariateKernel(data_kernel, num_tasks)
    k = mtkernel(x1, x2, params, noise)
    assert_equal(k.shape, (num_tasks*len(x1), num_tasks*len(x2)))


@pytest.mark.parametrize("data_kernel", [RBFKernel, MaternKernel])
@pytest.mark.parametrize("num_tasks", [2, 3])
@pytest.mark.parametrize("rank", [1, 2])
@pytest.mark.parametrize("dim", [1, 2])
def test_multivariate_kernel_shapes_test_noiseless(data_kernel, dim, num_tasks, rank):
    x1 = onp.random.randn(5, dim)
    x2 = onp.random.randn(3, dim)
    params = {"k_length": jnp.array(1.0), "k_scale": jnp.array(1.0),
              "W": jnp.ones((num_tasks, rank)), "v": jnp.ones(num_tasks)}
    mtkernel = MultivariateKernel(data_kernel, num_tasks)
    k = mtkernel(x1, x2, params)
    assert_equal(k.shape, (num_tasks*len(x1), num_tasks*len(x2)))


@pytest.mark.parametrize("data_kernel", [RBFKernel, MaternKernel])
@pytest.mark.parametrize("num_tasks", [2, 3])
@pytest.mark.parametrize("rank", [1, 2])
@pytest.mark.parametrize("dim", [1, 2])
def test_multivariate_kernel_shapes_train(data_kernel, dim, num_tasks, rank):
    x1 = onp.random.randn(5, dim)
    x2 = onp.random.randn(5, dim)
    params = {"k_length": jnp.array(1.0), "k_scale": jnp.array(1.0),
              "W": jnp.ones((num_tasks, rank)), "v": jnp.ones(num_tasks)}
    noise = jnp.ones(num_tasks)
    mtkernel = MultivariateKernel(data_kernel, num_tasks)
    k = mtkernel(x1, x2, params, noise)
    assert_equal(k.shape, (num_tasks*len(x1), num_tasks*len(x2)))


def test_LCMKernel():
    base_kernel = 'RBF'
    lcm_kernel = LCMKernel(base_kernel)
    assert_(callable(lcm_kernel), "The result of MultitaskKernel should be a function.")


@pytest.mark.parametrize("num_latent", [1, 2])
@pytest.mark.parametrize("data_kernel", [RBFKernel, MaternKernel])
@pytest.mark.parametrize("dim", [1, 2])
def test_LCMKernel_shapes_multitask(data_kernel, dim, num_latent):
    x1 = onp.random.randn(5, dim)
    x2 = onp.random.randn(3, dim)
    x1 = onp.column_stack([x1, onp.zeros_like(x1)])
    x2 = onp.column_stack([x2, onp.ones_like(x2)])
    x12 = onp.vstack([x1, x2])
    params = {"k_length": jnp.ones(num_latent), "k_scale": jnp.ones(num_latent),
              "W": jnp.ones((num_latent, 2, 2)), "v": jnp.ones((num_latent, 2))}
    noise = jnp.array([1.0, 1.0])
    mtkernel = LCMKernel(data_kernel, shared_input_space=False)
    k = mtkernel(x12, x12, params, noise)
    assert_equal(k.shape, (len(x12), len(x12)))


@pytest.mark.parametrize("num_latent", [1, 2])
@pytest.mark.parametrize("data_kernel", [RBFKernel, MaternKernel])
@pytest.mark.parametrize("num_tasks", [2, 3])
@pytest.mark.parametrize("rank", [1, 2])
@pytest.mark.parametrize("dim", [1, 2])
def test_LCMKernel_shapes_multivariate(data_kernel, dim, num_latent, rank, num_tasks):
    x1 = onp.random.randn(5, dim)
    x2 = onp.random.randn(3, dim)
    params = {"k_length": jnp.ones(num_latent), "k_scale": jnp.ones(num_latent),
              "W": jnp.ones((num_latent, num_tasks, rank)), "v": jnp.ones((num_latent, num_tasks))}
    noise = jnp.ones(num_tasks)
    mtkernel = LCMKernel(data_kernel, shared_input_space=True, num_tasks=num_tasks)
    k = mtkernel(x1, x2, params, noise)
    assert_equal(k.shape, (num_tasks*len(x1), num_tasks*len(x2)))



__author__ = "Maxim Ziatdinov"
__copyright__ = "Copyright Maxim Ziatdinov (2021)"
__maintainer__ = "Maxim Ziatdinov"
__email__ = "maxim.ziatdinov@ai4microcopy.com"

from setuptools import setup, find_packages
import os

module_dir = os.path.dirname(os.path.abspath(__file__))

with open(os.path.join(module_dir, 'gpax/__version__.py')) as f:
    __version__ = f.read().split("'")[1]

if __name__ == "__main__":
    setup(
        name='gpax',
        python_requires='>=3.9',
        version=__version__,
        description='Gaussian processes in NumPyro and JAX',
        long_description=open(os.path.join(module_dir, 'README.md')).read(),
        long_description_content_type='text/markdown',
        url='https://github.com/ziatdinovmax/gpax/',
        author='Maxim Ziatdinov',
        author_email='maxim.ziatdinov@ai4microcopy.com',
        license='MIT license',
        packages=find_packages(),
        zip_safe=False,
        install_requires=[
            'jax>=0.4.8',
            'numpyro>=0.11.0',
            'dm-haiku>=0.0.5',
            'matplotlib>=3.1'
        ],
        classifiers=['Programming Language :: Python',
                     'Development Status :: 3 - Alpha',
                     'Intended Audience :: Science/Research',
                     'Operating System :: POSIX :: Linux',
                     'Operating System :: MacOS :: MacOS X',
                     'Topic :: Scientific/Engineering']
    )



from functools import partial
from typing import Callable, Dict, Optional, Tuple

import jax
from jax import jit
import jax.numpy as jnp
import numpy as onp
import numpyro
import numpyro.distributions as dist
from numpyro.contrib.module import random_haiku_module

from ..kernels import LCMKernel
from .vidkl import viDKL


class viMTDKL(viDKL):
    """
    Implementation of the variational infernece-based deep kernel learning
    for multi-task/fidelity problems

    Args:
        input_dim:
            Number of input dimensions, not counting the column with task indices (if any)
        z_dim:
            Latent space dimensionality (defaults to 2)
        data_kernel:
            Kernel function operating on data inputs ('RBF', 'Matern', 'Periodic', or a custom function)
        num_latents:
            Number of latent functions. Typically equal to or less than the number of tasks
        shared_input_space:
            If True (default), assumes that all tasks share the same input space and
            uses a multivariate kernel (Kronecker product). If False, assumes that different tasks
            have different number of observations and uses a multitask kernel (elementwise multiplication).
            In that case, the task indices must be appended as the last column of the input vector.
        num_tasks:
            Number of tasks. This is only needed if `shared_input_space` is True.
        rank:
            Rank of the weight matrix in the task kernel. Cannot be larger than the number of tasks.
            Higher rank implies higher correlation. Uses *(num_tasks - 1)* when not specified.
        data_kernel_prior:
            Optional priors over kernel hyperparameters; uses LogNormal(0,1) by default
        nn:
            Custom neural network ('feature extractor'); uses a 3-layer MLP
            with ReLU activations by default
        latent_prior:
            Optional prior over the latent space (NN embedding); uses none by default
        guide:
            Auto-guide option, use 'delta' (default) or 'normal'
        W_prior_dist:
            Optional custom prior distribution over W in the task kernel, :math:`WW^T + diag(v)`.
            Defaults to Normal(0, 10).
        v_prior_dist:
            Optional custom prior distribution over v in the task kernel, :math:`WW^T + diag(v)`.
            Must be non-negative. Defaults to LogNormal(0, 1)
        task_kernel_prior:
            Optional custom priors over task kernel parameters;
            Defaults to Normal(0, 10) for weights W and LogNormal(0, 1) for variances v.

        **kwargs:
            Optional custom prior distributions over observational noise (noise_dist_prior)
            and kernel lengthscale (lengthscale_prior_dist)
    """

    def __init__(self, input_dim: int, z_dim: int = 2, data_kernel: str = 'RBF',
                 num_latents: int = None, shared_input_space: bool = True,
                 num_tasks: int = None, rank: Optional[int] = None,
                 data_kernel_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 nn: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None,
                 guide: str = 'delta',
                 W_prior_dist: Optional[dist.Distribution] = None,
                 v_prior_dist: Optional[dist.Distribution] = None,
                 task_kernel_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 **kwargs) -> None:
        args = (input_dim, z_dim, None, None, nn, None, guide)
        super(viMTDKL, self).__init__(*args, **kwargs)
        if shared_input_space:
            if num_tasks is None:
                raise ValueError("Please specify num_tasks")
        else:
            if num_latents is None:
                raise ValueError("Please specify num_latents")
        self.num_tasks = num_tasks
        self.num_latents = num_tasks if num_latents is None else num_latents
        self.rank = rank
        self.kernel = LCMKernel(
            data_kernel, shared_input_space, num_tasks, **kwargs)
        self.data_kernel_prior = data_kernel_prior
        self.task_kernel_prior = task_kernel_prior
        self.shared_input = shared_input_space
        self.W_prior_dist = W_prior_dist
        self.v_prior_dist = v_prior_dist

    def model(self, X: jnp.ndarray, y: jnp.ndarray = None, **kwargs) -> None:
        """Multitask DKL probabilistic model"""

        # Check that we have necessary info for sampling kernel params
        if not self.shared_input and self.num_tasks is None:
            self.num_tasks = len(onp.unique(self.X_train[:, -1]))

        if self.rank is None:
            self.rank = self.num_tasks - 1

        # NN part
        feature_extractor = random_haiku_module(
            "feature_extractor", self.nn_module, input_shape=(1, *self.data_dim),
            prior=(lambda name, shape: dist.Cauchy() if name.startswith("b") else dist.Normal()))
        z = feature_extractor(X if self.shared_input else X[:, :-1])
        if not self.shared_input:
            z = jnp.column_stack((z, X[:, -1]))

        # Initialize GP kernel mean function at zeros
        if self.shared_input:
            f_loc = jnp.zeros(self.num_tasks * X.shape[0])
        else:
            f_loc = jnp.zeros(X.shape[0])

        # Sample data kernel parameters
        if self.data_kernel_prior:
            data_kernel_params = self.data_kernel_prior()
        else:
            data_kernel_params = self._sample_kernel_params()

        # Sample task kernel parameters
        if self.task_kernel_prior:
            task_kernel_params = self.task_kernel_prior()
        else:
            task_kernel_params = self._sample_task_kernel_params()

        # Combine two dictionaries with parameters
        kernel_params = {**data_kernel_params, **task_kernel_params}

        # Sample noise
        if self.noise_prior:  # this will be removed in the future releases
            noise = self.noise_prior()
        else:
            noise = self._sample_noise()

        # Compute multitask_kernel
        k = self.kernel(z, z, kernel_params, noise, **kwargs)

        # Sample y according to the standard Gaussian process formula
        numpyro.sample(
            "y",
            dist.MultivariateNormal(loc=f_loc, covariance_matrix=k),
            obs=y,
        )

    def _sample_noise(self):
        """Sample observational noise"""
        if self.noise_prior_dist is not None:
            noise_dist = self.noise_prior_dist
        else:
            noise_dist = dist.LogNormal(
                    jnp.zeros(self.num_tasks),
                    jnp.ones(self.num_tasks))

        noise = numpyro.sample("noise", noise_dist.to_event(1))
        return noise

    def _sample_task_kernel_params(self):
        """
        Sample task kernel parameters with default weakly-informative priors
        or custom priors for all the latent functions
        """
        if self.W_prior_dist is not None:
            W_dist = self.W_prior_dist
        else:
            W_dist = dist.Normal(
                    jnp.zeros(shape=(self.num_latents, self.num_tasks, self.rank)),  # loc
                    10*jnp.ones(shape=(self.num_latents, self.num_tasks, self.rank)) # var
            )
        if self.v_prior_dist is not None:
            v_dist = self.v_prior_dist
        else:
            v_dist = dist.LogNormal(
                    jnp.zeros(shape=(self.num_latents, self.num_tasks)),  # loc
                    jnp.ones(shape=(self.num_latents, self.num_tasks)) # var
            )
        with numpyro.plate("latent_plate_task", self.num_latents):
            W = numpyro.sample("W", W_dist.to_event(2))
            v = numpyro.sample("v", v_dist.to_event(1))
        return {"W": W, "v": v}

    def _sample_kernel_params(self):
        """
        Sample data ("base") kernel parameters with default weakly-informative
        priors for all the latent functions
        """
        squeezer = lambda x: x.squeeze() if self.num_latents > 1 else x
        with numpyro.plate("latent_plate_data", self.num_latents, dim=-2):
            with numpyro.plate("ard", self.kernel_dim, dim=-1):
                length = numpyro.sample("k_length", dist.LogNormal(0.0, 1.0))
            scale = numpyro.sample("k_scale", dist.Normal(1.0, 1e-4))
        return {"k_length": squeezer(length), "k_scale": squeezer(scale)}

    #@partial(jit, static_argnames='self')
    def get_mvn_posterior(self,
                          X_new: jnp.ndarray,
                          nn_params: Dict[str, jnp.ndarray],
                          k_params: Dict[str, jnp.ndarray],
                          noiseless: bool = False,
                          y_residual: jnp.ndarray = None,
                          **kwargs
                          ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """
        Returns predictive mean and covariance at new points
        (mean and cov, where cov.diagonal() is 'uncertainty')
        given a single set of DKL parameters
        """
        if y_residual is None:
            y_residual = self.y_train
        noise = k_params["noise"]
        noise_p = noise * (1 - jnp.array(noiseless, int))
        # embed data into the latent space
        z_train = self.nn_module.apply(
            nn_params, jax.random.PRNGKey(0),
            self.X_train if self.shared_input else self.X_train[:, :-1])
        z_test = self.nn_module.apply(
            nn_params, jax.random.PRNGKey(0),
            X_new if self.shared_input else X_new[:, :-1])
        if not self.shared_input:
            z_train = jnp.column_stack((z_train, self.X_train[:, -1]))
            z_test = jnp.column_stack((z_test, X_new[:, -1]))
        # compute kernel matrices for train and test data
        k_pp = self.kernel(z_test, z_test, k_params, noise_p, **kwargs)
        k_pX = self.kernel(z_test, z_train, k_params, jitter=0.0)
        k_XX = self.kernel(z_train, z_train, k_params, noise, **kwargs)
        # compute the predictive covariance and mean
        K_xx_inv = jnp.linalg.inv(k_XX)
        cov = k_pp - jnp.matmul(k_pX, jnp.matmul(K_xx_inv, jnp.transpose(k_pX)))
        mean = jnp.matmul(k_pX, jnp.matmul(K_xx_inv, y_residual))
        return mean, cov



from typing import Optional
import jax
import jax.numpy as jnp


def compute_penalty(X: jnp.ndarray, recent_points: jnp.ndarray,
                    penalty_type: str = "delta", penalty_factor: float = 1.0):
    """
    Compute penalties for points in X for Bayesian Optimization's acquisition function 
    based on a specified penalty type.

    Args:

        X: The array of points in the domain of the objective function for which to compute penalties.
        recent_points: The array of recently evaluated points to consider when computing penalties.
        penalty_type: The type of penalty to apply. Must be either 'delta' or 'inverse_distance'.
        Defaults to 'delta'.
        penalty_factor: The constant multiplier for the 'inverse_distance' penalty. Ignored if penalty_type is 'delta'.
        Defaults to 1.

    Returns:

        The array of computed penalties for each point in X, to be applied to
        the acquisition function values.
    """
    p = penalty_factor
    if penalty_type not in ["delta", "inverse_distance", "inverse distance"]:
        raise NotImplementedError(
            "Avaialble penalty types are 'delta' and 'inverse distance'")
    if penalty_type == "delta":
        penalties = find_and_replace_point_indices(X, recent_points)
    else:
        penalties = p * jax.vmap(penalty_point, in_axes=(0, None))(X, recent_points)
    return penalties


def penalty_point(x: jnp.ndarray, recent_points: jnp.ndarray) -> jnp.ndarray:
    """
    Compute a penalty for point x based on its distance to recent points.
    """
    if recent_points.ndim == 1:
        recent_points = recent_points[:, None] 
    distances = jnp.linalg.norm(recent_points - x, axis=1)
    # Penalties are inversely proportional to distance and timestamp
    if len(recent_points) == 1:
        timestamps = 1
    else:
        timestamps = jnp.arange(len(recent_points) + 1, 1, -1)
    penalties = 1 / (distances + 1) / timestamps
    return jnp.sum(penalties)


def find_and_replace_point_indices(points, other_points):
    # Create an array of zeros
    zero_array = jnp.zeros(len(points))

    for single_point in other_points:
        # Check where in points each element of single_point exists
        index = jnp.where(jnp.all(points == single_point, axis=1))

        # If index array is not empty, replace corresponding zero_array element with jnp.inf
        if index[0].size > 0:
            zero_array = zero_array.at[index[0][0]].set(jnp.inf)

    # Return the modified array
    return zero_array



from typing import Callable, Dict, Optional

import jax.numpy as jnp
import numpy as onp
import numpyro
import numpyro.distributions as dist

from .gp import ExactGP
from ..kernels import MultitaskKernel


class CoregGP(ExactGP):

    """
    Coregionalized Gaussian Process model

    Args:
        input_dim:
            Number of input dimensions
        data_kernel:
            Kernel function operating on data inputs ('RBF', 'Matern', 'Periodic', or a custom function)
        mean_fn:
            Optional deterministic mean function (use 'mean_fn_priors' to make it probabilistic)
        data_kernel_prior:
            Optional custom priors over the data kernel hyperparameters; uses LogNormal(0,1) by default
        mean_fn_prior:
            Optional priors over mean function parameters
        noise_prior:
            Optional custom prior for observation noise; uses LogNormal(0,1) by default.
        task_kernel_prior:
            Optional custom priors over task kernel parameters;
            Defaults to Normal(0, 10) for weights W and LogNormal(0, 1) for variances v.
        rank: int
            Rank of the weight matrix in the task kernel. Cannot be larger than the number of tasks.
            Higher rank implies higher correlation. Defaults to 1.

    """
    def __init__(self, input_dim: int, data_kernel: str,
                 mean_fn: Optional[Callable[[jnp.ndarray, Dict[str, jnp.ndarray]], jnp.ndarray]] = None,
                 data_kernel_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 mean_fn_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 noise_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 task_kernel_prior: Optional[Callable[[], Dict[str, jnp.ndarray]]] = None,
                 rank: int = 1, **kwargs) -> None:
        args = (input_dim, None, mean_fn, None, mean_fn_prior, noise_prior)
        super(CoregGP, self).__init__(*args)
        self.num_tasks = None
        self.rank = rank
        self.kernel = MultitaskKernel(data_kernel, **kwargs)
        self.data_kernel_prior = data_kernel_prior
        self.task_kernel_prior = task_kernel_prior
        self.kernel_name = data_kernel

    def model(self,
              X: jnp.ndarray,
              y: jnp.ndarray = None,
              **kwargs: float
              ) -> None:
        """Multitask GP probabilistic model with inputs X and targets y"""
        self.num_tasks = len(onp.unique(X[:, -1]))
        # Initialize mean function at zeros
        f_loc = jnp.zeros(X.shape[0])

        # Sample data kernel parameters
        if self.data_kernel_prior:
            data_kernel_params = self.data_kernel_prior()
        else:
            data_kernel_params = self._sample_kernel_params(output_scale=False)

        # Sample task kernel parameters
        if self.task_kernel_prior:
            task_kernel_params = self.task_kernel_prior()
        else:
            task_kernel_params = self._sample_task_kernel_params(self.num_tasks, self.rank)

        # Combine two dictionaries with parameters
        kernel_params = {**data_kernel_params, **task_kernel_params}

        # Sample noise
        if self.noise_prior:
            noise = self.noise_prior()
        else:  # consider using numpyro.plate here
            noise = numpyro.sample(
                "noise", dist.LogNormal(
                    jnp.zeros(self.num_tasks), jnp.ones(self.num_tasks))
            )

        # Compute multitask_kernel
        k = self.kernel(X, X, kernel_params, noise)

        # Add mean function (if any)
        if self.mean_fn is not None:
            args = [X]
            if self.mean_fn_prior is not None:
                args += [self.mean_fn_prior()]
            f_loc += self.mean_fn(*args).squeeze()

        # sample y according to the standard Gaussian process formula
        numpyro.sample(
            "y",
            dist.MultivariateNormal(loc=f_loc, covariance_matrix=k),
            obs=y,
        )

    def _sample_task_kernel_params(self, n_tasks, rank):
        """
        Sample task kernel parameters with default weakly-informative priors
        """
        W = numpyro.sample("W", numpyro.distributions.Normal(
            jnp.zeros(shape=(n_tasks, rank)), 10*jnp.ones(shape=(n_tasks, rank))))
        v = numpyro.sample("v", numpyro.distributions.LogNormal(
            jnp.zeros(shape=(n_tasks,)), jnp.ones(shape=(n_tasks,))))
        return {"W": W, "v": v}



from .gp import ExactGP
from .vgp import vExactGP
from .vigp import viGP
from .spm import sPM
from .ibnn import iBNN
from .vi_ibnn import vi_iBNN
from .dkl import DKL
from .vidkl import viDKL
from .vi_mtdkl import viMTDKL
from .mtgp import MultiTaskGP
from .corgp import CoregGP

__all__ = [
    "ExactGP",
    "vExactGP",
    "viGP",
    "sPM",
    "iBNN",
    "vi_iBNN",
    "DKL",
    "viDKL",
    "viMTDKL",
    "MultiTaskGP",
    "CoregGP"
]
